from dataclasses import dataclass
from typing import List, Optional

from ego4d.research.common import SlurmConfig


@dataclass
class InputConfig:
    feature_path: str
    metadata_path: str
    features_per_second: float
    narration_width_sample_sec: float
    max_num_feature_vec_video_uids: int
    dsets_to_use: List[str]


@dataclass
class ModelConfig:
    nlp_feature_size: int
    visual_feature_size: int
    final_proj_size: int


@dataclass
class K400PreprocessConfig:
    dataset_dir: str
    set_to_use: str
    root_dir: str
    viz_feature_path: str
    metadata_out_path: str
    feature_extract_config_path: str
    num_labels_per_machine: int


@dataclass
class EgoPreprocessFeatureConfig:
    hdf5_path: str


@dataclass
class EgoPreprocessNarrConfig:
    narration_json_path: str
    num_workers: int
    st_model_name: str
    accelerator: str
    root_dir: str
    metadata_out_path: str
    narration_out_dir: str
    limit: Optional[int]
    num_narrs_per_machine: int


@dataclass
class EgoCharadePreprocessConfig:
    set_path: str
    video_root_path: str
    class_desc_path: str
    out_path: str
    out_label_path: str
    num_vids_per_machine: int


@dataclass
class CCPreprocessConfig:
    in_path: str
    helper_workers: int

    hdf5_viz_path: str
    hdf5_sent_path: str
    meta_path: str
    batch_size: int
    num_workers: int
    prefetch_factor: int
    imgs_per_gpu: int


@dataclass
class PreprocessConfig:
    slurm_config: SlurmConfig

    mode: str
    root_dir: str
    ego4d_narr: EgoPreprocessNarrConfig
    ego4d_features: EgoPreprocessFeatureConfig
    k400: K400PreprocessConfig
    ego_charade: EgoCharadePreprocessConfig
    cc: CCPreprocessConfig


@dataclass
class TrainConfig:
    input_config: InputConfig
    model_config: ModelConfig
    pre_config: PreprocessConfig

    checkpoint_dir: str
    checkpoint_metric: str
    batch_size: int
    num_workers: int
    prefetch_factor: int

    num_epochs: int
    accelerator: str
    devices: int

    run_locally: bool
    tb_log_dir: str
    tb_log_name: str

    lr: float
    beta1: float
    beta2: float
    wd: float
    eps: float

    eval_per_iter: int
    eval_init: bool

    use_soft_loss: Optional[bool]
    soft_loss_threshold: float
    use_bce: bool
    use_logit_scale: bool
    norm_logits: bool



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

"""
Functionality related to parsing the video manifest and storing an in-memory
representation for the download operation.
"""
import csv
import datetime
import logging
import os
import tempfile
from pathlib import Path
from typing import Dict, Iterable, Set

import regex

from ego4d.cli.s3path import bucket_and_key_from_path
from ego4d.cli.universities import BUCKET_TO_UNIV

__MANIFEST_BUCKET = "ego4d-consortium-sharing"
__METADATA_FILENAME = "ego4d.json"
__DATASETS_FILENAME = "datasets.csv"


class VideoMetadata:
    """
    Data object that corresponds to a single video entry in a manifest CSV file.
    """

    __FILE_UID_KEY = "file_uid"
    __VIDEO_UID_KEY = "video_uid"
    __S3_LOCATION_KEYS = ["canonical_s3_location", "s3_path"]
    __FILE_TYPE_KEY = "type"
    __BENCHMARKS_KEY = "benchmarks"

    def __init__(self, row: Dict[str, str]):
        # The raw contents of the CSV row
        self.raw_data: Dict[str, str] = dict(row)

        # Unique identifier for the video
        if self.__FILE_UID_KEY in row:
            self.file_download = True
            self.uid: str = row[self.__FILE_UID_KEY]
        else:
            assert (
                self.__VIDEO_UID_KEY in row
            ), "Either file_uid or video_uid must be specified"
            self.file_download = False
            self.uid: str = row[self.__VIDEO_UID_KEY]

        # Path to the video file on AWS S3 (e.g. "s3://bucket/key")
        for x in self.__S3_LOCATION_KEYS:
            if x in row:
                self.s3_path: str = row[x]
                break

        # Name of the S3 bucket that holds the video
        self.s3_bucket: str = None

        # S3 object key for the video
        self.s3_object_key: str = None

        if self.s3_path:
            self.s3_bucket, self.s3_object_key = bucket_and_key_from_path(self.s3_path)
            self.university: str = BUCKET_TO_UNIV.get(self.s3_bucket, "")

        type = row.get(self.__FILE_TYPE_KEY)
        if type in ["mp4", "video"]:
            self.file_download = False
        elif type in ["file", "json"]:
            self.file_download = True
        else:
            # Default to above
            pass

        if self.file_download:
            self.filename_base = os.path.basename(self.s3_path)
        else:
            self.filename_base = f"{self.uid}.mp4"

        benchmarks = row.get(self.__BENCHMARKS_KEY)
        if benchmarks:
            self.benchmarks = regex.sub(r"\s+", "", benchmarks.lower())
        else:
            self.benchmarks = None


def list_videos_in_manifest(
    manifest_path: Path, benchmarks: Set[str], for_universities: Set[str]
) -> Iterable[VideoMetadata]:
    """
    Creates a generator that reads every row of a manifest CSV file and returns the row
    as a VideoMetadata object.

    Args:
        manifest_path: Path on local disk to a manifest CSV file
        for_universities: Only videos belonging to universities in this set will be
            returned. If the set is empty then all videos will be returned.
    """
    with open(manifest_path, newline="") as f:
        reader = csv.DictReader(f)

        has_benchmarks = False
        if len(benchmarks) > 0:
            if "benchmarks" in reader.fieldnames:
                has_benchmarks = True
                benchmarks = [x.lower() for x in benchmarks]
                b_re = regex.compile(r"\[(\w+)?(?:\|(\w+))*\]", regex.IGNORECASE)
                print(f"Filtering by benchmarks: {benchmarks}")
            else:
                print(
                    "Benchmarks specified but ignored without a benchmarks field in manifest."
                )

        for row in reader:
            metadata = VideoMetadata(row)
            if has_benchmarks:
                if not metadata.benchmarks:
                    continue
                m = b_re.match(metadata.benchmarks)
                if not m:
                    if metadata.benchmarks:
                        logging.warning(
                            f"Invalid benchmarks manifest entry ignored: {metadata.benchmarks}"
                        )
                    continue
                grps = m.captures(1) + m.captures(2)
                cnt_bars = metadata.benchmarks.count("|")
                if cnt_bars > 0:
                    assert len(grps) == (
                        cnt_bars + 1
                    ), f"Invalid benchmarks row: {metadata.benchmarks}"
                else:
                    assert (
                        len(grps) <= 1
                    ), f"Invalid benchmarks row: {metadata.benchmarks}"
                if not any(x in benchmarks for x in grps):
                    continue
            if for_universities and metadata.university not in for_universities:
                continue
            yield metadata


def download_manifest_for_version(
    version: str, dataset: str, download_dir: Path, s3
) -> Path:
    """
    Downloads the manifest file to the download_path as a file named "manifest.csv"

    Args:
        version:
        dataset:
        download_dir:
        s3 (S3.ServiceResource):

    Returns:
    """
    download_path = download_dir / "manifest.csv"
    _manifest_object(version, dataset, s3).download_file(str(download_path))
    return download_path


def _manifest_object(version: str, dataset: str, s3):
    """

    Args:
        version:
        dataset:
        s3 (S3.ServiceResource):

    Returns:
    S3.Object
    """
    return s3.Bucket(__MANIFEST_BUCKET).Object(
        f"public/{version}/{dataset}/manifest.csv"
    )


def _metadata_object(version: str, s3):
    """
    The primary metadata JSON
    """
    return s3.Bucket(__MANIFEST_BUCKET).Object(
        f"public/{version}/{__METADATA_FILENAME}"
    )


def download_metadata(version: str, download_dir: Path, s3) -> Path:
    """
    Downloads the primary metadata JSON to the download_path
    """
    download_path = download_dir / __METADATA_FILENAME
    if download_path.exists():
        # TODO: Check for file version
        return download_path
    else:
        print("Downloading Ego4D metadata json..")

    _metadata_object(version, s3).download_file(str(download_path))
    return download_path


def _datasets_object(version: str, s3):
    """
    The primary metadata JSON
    """
    return s3.Bucket(__MANIFEST_BUCKET).Object(
        f"public/{version}/{__DATASETS_FILENAME}"
    )


def download_datasets(version: str, download_dir: Path, s3) -> Path:
    """
    Downloads the primary datasets csv to the download_path
    """
    download_path = download_dir / __DATASETS_FILENAME
    if download_path.exists():
        mtime = datetime.datetime.fromtimestamp(
            download_path.stat().st_mtime, tz=datetime.timezone.utc
        )
        delta = (datetime.utcnow() - mtime).totalseconds() / 3600
        if delta < 12:
            print("Bypassing recent datasets.csv..")
            return download_path
    else:
        print("Downloading datasets.csv..")

    _datasets_object(version, s3).download_file(str(download_path))
    return download_path


def print_datasets(version: str, s3) -> None:
    assert version
    try:
        with tempfile.TemporaryDirectory() as tmpdir:
            tmppath = Path(tmpdir)
            p = download_datasets(version, tmppath, s3)
            if not p.exists():
                logging.error("datasets.csv download failed!")
                print("Download datasets.csv error (defaulting to local)..")
                p = Path(__file__).with_name("datasets.csv")
            with p.open("r") as f:
                rows = csv.DictReader(f)
                print("\nAvailable Ego4D datasets:")
                for row in rows:
                    print(f"   {row['dataset']:<21}\t{row['description']}")
                print()
    except Exception as ex:
        logging.exception(f"Exception retrieving Ego4D datasets: {ex}")



# Copyright (c) Meta Platforms, Inc. and affiliates. All Rights Reserved.

from dataclasses import dataclass
from typing import Optional, Tuple

import torch
from ego4d.features.config import BaseModelConfig, InferenceConfig
from ego4d.features.models.common import FeedVideoInput
from pytorchvideo.transforms import ApplyTransformToKey, ShortSideScale
from torch.nn import Identity, Module
from torchvision.transforms import Compose, Lambda
from torchvision.transforms._transforms_video import CenterCropVideo, NormalizeVideo


@dataclass
class ModelConfig(BaseModelConfig):
    model_path: Optional[str] = None
    hub_path: Optional[str] = "slowfast_r101"
    slowfast_alpha: int = 4

    # transformation config
    side_size: int = 256
    crop_size: int = 256
    mean: Tuple[float] = (0.45, 0.45, 0.45)
    std: Tuple[float] = (0.225, 0.225, 0.225)


class GetFv(Module):
    def __init__(self):
        super().__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        bs = x.shape[0]
        fv_s = x.shape[1]
        return x.view(bs, fv_s, -1).mean(2)


def load_model(
    inference_config: InferenceConfig,
    config: ModelConfig,
    patch_final_layer: bool = True,
) -> Module:
    if config.model_path is not None:
        raise AssertionError("not supported yet")
        model = None
    else:
        assert config.hub_path is not None
        model = torch.hub.load(
            "facebookresearch/pytorchvideo", config.hub_path, pretrained=True
        )

    assert model is not None

    if patch_final_layer:
        model.blocks[6] = GetFv()

    # Set to GPU or CPU
    model = FeedVideoInput(model)
    model = model.eval()
    model = model.to(inference_config.device)
    return model


class PackPathway(torch.nn.Module):
    """
    Transform for converting video frames as a list of tensors.
    """

    def __init__(self, slow_fast_alpha):
        super().__init__()
        self.slow_fast_alpha = slow_fast_alpha

    def forward(self, frames: torch.Tensor):
        fast_pathway = frames
        # Perform temporal sampling from the fast pathway.
        slow_pathway = torch.index_select(
            frames,
            1,
            torch.linspace(
                0, frames.shape[1] - 1, frames.shape[1] // self.slow_fast_alpha
            ).long(),
        )
        frame_list = [slow_pathway, fast_pathway]
        return frame_list


def get_transform(inference_config: InferenceConfig, config: ModelConfig):
    return ApplyTransformToKey(
        key="video",
        transform=Compose(
            [
                Lambda(lambda x: x / 255.0),
                NormalizeVideo(config.mean, config.std),
                ShortSideScale(size=config.side_size),
                CenterCropVideo(config.crop_size),
                PackPathway(config.slowfast_alpha),
            ]
        ),
    )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

import argparse
import datetime
import functools
import math
import os
import sys
from dataclasses import dataclass
from typing import List, Optional, Tuple

import hydra

import numpy as np
import submitit
import torch
from ego4d.features.config import (
    FeatureExtractConfig,
    get_videos,
    ScheduleConfig,
    Video,
)
from ego4d.features.extract_features import num_fvs, perform_feature_extraction
from hydra.core.config_store import ConfigStore
from omegaconf import OmegaConf
from tqdm import tqdm


def greedy_create_batches(
    videos: List[Video], times: List[float], max_time_per_batch: int
) -> List[List[Tuple[Video, float]]]:
    """
    Greedy bin packing algorithm.

    For each bin:
    - Packs the most amount of big videos into the bin
    - Pack any small videos remaining that fit

    big/small corresponding to the time estimate for completion
    """
    assert len(videos) > 0, "empty videos"
    assert len(videos) == len(times), f"videos={len(videos)} vs. times={len(times)}"

    vt = [(v, t) for v, t in zip(videos, times)]
    vt.sort(key=lambda x: x[1])

    assert (
        vt[-1][1] <= max_time_per_batch
    ), f"""
        can't batch things if the max time is larger than max_time_per_batch
        {vt[-1][-1]} vs. {max_time_per_batch}
    """

    n = len(videos)

    i = 0
    j = n - 1

    batches = []
    curr = []
    curr_time = 0

    while i <= j:
        old_i, old_j = i, j

        while i <= j and curr_time + times[i] <= max_time_per_batch:
            curr.append((videos[i], times[i]))
            curr_time += times[i]
            i += 1

        while j >= i and curr_time + times[j] <= max_time_per_batch:
            curr.append((videos[j], times[j]))
            curr_time += times[j]
            j -= 1

        assert (
            i != old_i or j != old_j
        ), f"""
        Could not batch it up -
            i = {i}, j = {j}
            old_i = {old_i}, old_j = {old_j}
            time_i = {times[i]}, time_j = {times[j]}
            max_time_per_batch = {max_time_per_batch}
        """

        batches.append(curr)
        curr = []
        curr_time = 0

    if len(curr) > 0:
        batches.append(curr)

    return batches


def validate_batches(timeout_minutes, times_per_batch, batched_vids):
    for t, b in zip(times_per_batch, batched_vids):
        print(len(b), t)
        assert (
            t <= timeout_minutes * 60
        ), f"""
            Algorithm to batch videos is incorrect.

            Batch estimated time = {t}
            Timeout minutes = {timeout_minutes}
            Timeout seconds = {timeout_minutes * 60}
        """


def batch_videos(
    videos: List[Video], config: FeatureExtractConfig
) -> List[List[Video]]:
    # estimate each time to extract per sub-clip
    num_forward_passes = [
        np.ceil(
            num_fvs(v, config.inference_config) / config.inference_config.batch_size
        )
        for v in videos
    ]
    times = [
        config.schedule_config.overhead
        * n
        * config.schedule_config.time_per_forward_pass
        for n in num_forward_passes
    ]

    # batch up these videos into config.timeout_min batches
    batched_vt = greedy_create_batches(
        videos, times, max_time_per_batch=config.schedule_config.timeout_min * 60
    )
    batched_vids = [[uid for uid, _ in b] for b in batched_vt]

    times_per_batch = [sum(t for _, t in b) for b in batched_vt]

    indicies = list(range(len(times_per_batch)))
    indicies.sort(key=lambda i: times_per_batch[i])
    batched_vids = [batched_vids[i] for i in indicies]
    times_per_batch = [times_per_batch[i] for i in indicies]

    validate_batches(config.schedule_config.timeout_min, times_per_batch, batched_vids)
    return batched_vids


def print_stats_for_videos(
    config: FeatureExtractConfig, all_videos: List[Video], videos: List[Video]
):
    # stats for uids
    assert isinstance(all_videos[0], Video)
    assert isinstance(videos[0], Video)
    total_secs_uncompleted = sum(v.frame_count * config.fps for v in videos)
    secs_uncompleted = sum(v.frame_count * config.fps for v in all_videos)

    print(
        f"""
    Total Number of Videos = {len(all_videos)}
    Incomplete videos = {len(videos)}

    Total Seconds = {total_secs_uncompleted}
    Incomplete seconds = {secs_uncompleted} = {secs_uncompleted/total_secs_uncompleted * 100:.2f}%
    """
    )


def print_stats_for_scheduling(
    config: FeatureExtractConfig, batch_vids: List[List[Video]]
):
    schedule_time = config.schedule_config.schedule_time_per_node

    n_schedules = math.ceil(
        len(batch_vids) / config.schedule_config.slurm_array_parallelism
    )
    print("n schedules =", n_schedules)
    schedule_overhead = n_schedules * schedule_time * 60

    print(f"Will schedule {n_schedules}")

    sec_to_take = (
        n_schedules * 60 * config.schedule_config.timeout_min + schedule_overhead
    )
    print(sec_to_take)
    print(
        f"""
    {len(batch_vids)} batches
    {config.schedule_config.slurm_array_parallelism} batch of machines.

    Will take: {sec_to_take} seconds
        Schedule overhead: - {schedule_overhead} ({100*schedule_overhead/sec_to_take:.2f}%)
    """
    )


def print_completion_stats(results):
    time_to_load = []
    time_to_transfer = []
    forward_pass_time = []
    print(
        "overall,save_time,avg_total,avg_load_time,avg_transfer_time,avg_forward_pass"
    )
    for result in results:
        time_to_load.extend(result.to_load)
        time_to_transfer.extend(result.transfer_device)
        forward_pass_time.extend(result.forward_pass)

        ttl = torch.Tensor(result.to_load)
        ttt = torch.Tensor(result.transfer_device)
        fpt = torch.Tensor(result.forward_pass)
        mean_sum = ttl.mean() + ttt.mean() + fpt.mean()
        print(
            f"{result.overall},{result.to_save},{mean_sum},{ttl.mean()},{ttt.mean()},{fpt.mean()}"
        )

    print("")
    print("")

    print("Averages")
    print("mean_forward,only_forward_pass,time_to_load,time_to_transfer")
    ttl = torch.Tensor(time_to_load)
    ttt = torch.Tensor(time_to_transfer)
    fpt = torch.Tensor(forward_pass_time)
    mean_sum = ttl.mean() + ttt.mean() + fpt.mean()
    print(f"{mean_sum},{ttl.mean()},{ttt.mean()},{fpt.mean()}")


def create_executor(config: ScheduleConfig):
    if not config.run_locally:
        print("Using slurm/auto executor")
        executor = submitit.AutoExecutor(folder=config.log_folder)
    else:
        print("Using local executor")
        executor = submitit.LocalExecutor(folder=config.log_folder)

    executor.update_parameters(
        timeout_min=config.timeout_min,
        constraint=config.constraint,
        slurm_partition=config.slurm_partition,
        slurm_array_parallelism=config.slurm_array_parallelism,
        gpus_per_node=config.gpus_per_node,
        cpus_per_task=config.cpus_per_task,
    )
    return executor


@hydra.main(config_path="configs", config_name=None)
def schedule_feature_extraction(config: FeatureExtractConfig):
    print("###################### Feature Extraction Config ####################")
    print(OmegaConf.to_yaml(config))
    print("############################################################")

    # Get uids + {uids -> duration}
    videos, all_videos = get_videos(config)
    os.makedirs(config.io.out_path, exist_ok=True)
    with open(f"{config.io.out_path}/config.yaml", "w") as out_f:
        out_f.write(OmegaConf.to_yaml(config))

    print_stats_for_videos(config, all_videos=all_videos, videos=videos)

    # ------ Compute real timeout per batch
    batch_vids = batch_videos(videos=videos, config=config)
    config.schedule_config.slurm_array_parallelism = min(
        config.schedule_config.slurm_array_parallelism, len(batch_vids)
    )

    # ------ Schedule
    executor = create_executor(config.schedule_config)
    print_stats_for_scheduling(config, batch_vids)

    if not config.force_yes:
        print(f"Time is: {datetime.datetime.now()}")
        cont = input("Continue? [y/N]: ")
        if cont != "y":
            print("Exiting...")
            sys.exit(0)

    jobs = executor.map_array(
        functools.partial(perform_feature_extraction, config=config),
        batch_vids,
    )
    print(f"Jobs: {jobs}")

    # TODO: display better stats in tqdm ?
    results = []
    for job in tqdm(jobs):
        results.append(job.result())
    print_completion_stats(results)


if __name__ == "__main__":
    schedule_feature_extraction()



"""
This scripts generates the listing of files that the CLI will download.

It does so by partitioning a list of ManifestEntry's into parts. See the
`manifests` dict below; the keys are the names of each part of the dataset.

This should only be used internally by the Eng Team of the EgoExo dataset.
"""

import json
import os
from collections import defaultdict

from concurrent.futures import ThreadPoolExecutor

from ego4d.internal.download.manifest import (
    manifest_dumps,
    manifest_loads,
    ManifestEntry,
    PathSpecification,
)

from ego4d.internal.s3 import S3Downloader
from iopath.common.file_io import PathManager
from iopath.common.s3 import S3PathHandler
from tqdm.auto import tqdm


def init_workers():
    pass


def local_map(xs, map_fn, num_workers):  # pyre-ignore
    with ThreadPoolExecutor(num_workers, initializer=init_workers) as pool:
        for x in tqdm(pool.map(map_fn, xs), total=len(xs)):
            yield x


pathmgr = PathManager()  # for downloading files
pathmgr.register_handler(S3PathHandler(profile="default"))

dev_release = False

if dev_release:
    ver_name = "dev"
    base_release_dir = "egoexo/releases/dev/"
    release_dir = f"s3://ego4d-consortium-sharing/{base_release_dir}"
    internal_release_dir = release_dir
else:
    ver_name = "v2"
    base_release_dir = f"egoexo-public/{ver_name}/"
    release_dir = f"s3://ego4d-consortium-sharing/{base_release_dir}"
    internal_release_dir = (
        f"s3://ego4d-consortium-sharing/egoexo/releases/public_internal/{ver_name}/"
    )


manifests = {
    "metadata": [],
    "annotations": [],
    "takes": [],
    "take_trajectory": [],
    "take_eye_gaze": [],
    "take_point_cloud": [],
    "take_vrs": [],
    "take_vrs_noimagestream": [],
    "captures": [],
    "capture_trajectory": [],
    "capture_eye_gaze": [],
    "capture_point_cloud": [],
    "downscaled_takes/448": [],
    "features/omnivore_video": [],
    "features/maws_clip_2b": [],
    "ego_pose_pseudo_gt": [],
    "expert_commentary": [],
    "take_transcription": [],
    "take_audio": [],
}
if dev_release:
    manifests["takes_dropped"] = []

egoexo = {
    "released_takes": os.path.join(internal_release_dir, "released_takes.json"),
    "takes": os.path.join(internal_release_dir, "takes.json"),
    "takes_dropped": os.path.join(internal_release_dir, "_takes_dropped.json"),
    "captures": os.path.join(internal_release_dir, "captures.json"),
    "physical_setting": os.path.join(internal_release_dir, "physical_setting.json"),
    "participants": os.path.join(internal_release_dir, "participants.json"),
    "visual_objects": os.path.join(internal_release_dir, "visual_objects.json"),
    "metadata": os.path.join(internal_release_dir, "metadata.json"),
}

downloader = S3Downloader("default")
print("Downloading metadata")
for k, out_path in tqdm(egoexo.items()):
    s3_path = os.path.join(release_dir, f"{k}.json")
    metadata = downloader.file_desc(s3_path)
    # if not pathmgr.exists(s3_path):
    if metadata is None:
        print(f"WARN: {s3_path} does not exist")
        continue
    # print(s3_path, out_path)
    paths = [
        PathSpecification(
            s3_path,
            f"{k}.json",
            size=metadata.size,
        )
    ]
    manifests["metadata"].append(
        ManifestEntry(
            uid=k,
            paths=paths,
        )
    )

for k, v in egoexo.items():
    if pathmgr.exists(v):
        egoexo[k] = json.load(pathmgr.open(v))
    else:
        egoexo[k] = []

if len(egoexo["released_takes"]) == 0:
    egoexo["released_takes"] = [x["take_uid"] for x in egoexo["takes"]]

s3_buckets = set()
for c in tqdm(egoexo["captures"]):
    s3_bucket = c["_s3_root_dir"].split("/")[2]
    s3_buckets.add(s3_bucket)

take_by_take_uid = {x["take_uid"]: x for x in egoexo["takes"] + egoexo["takes_dropped"]}
take_name_to_uid = {
    t["take_name"]: t["take_uid"] for t in (egoexo["takes"] + egoexo["takes_dropped"])
}

capture_cam_id_to_is_ego = {
    (c["capture_uid"], cam["cam_id"]): cam["is_ego"]
    for c in egoexo["captures"]
    for cam in c["cameras"]
}

# split_file_path = os.path.join(release_dir, "annotations/splits.json")
split_file_path = (
    "s3://ego4d-consortium-sharing/egoexo-public/v2/annotations/splits.json"
)
split_data = json.load(pathmgr.open(split_file_path))
take_uid_to_splits = {k: [v] for k, v in split_data["take_uid_to_split"].items()}

capture_uid_to_splits = defaultdict(set)
for take_uid, split in split_data["take_uid_to_split"].items():
    if take_uid not in take_by_take_uid:
        print("Missing", take_uid)
        continue
    capture_uid = take_by_take_uid[take_uid]["capture"]["capture_uid"]
    capture_uid_to_splits[capture_uid].add(split.lower())

take_uid_to_benchmarks = defaultdict(list)
take_uid_to_benchmarks.update(split_data.get("take_uid_to_benchmark", {}))

s3_buckets = {
    x["university_id"]
    for x in egoexo["takes"]
    if x["take_uid"] in egoexo["released_takes"]
}

manifests["downscaled_takes/448"] = []
if "downscaled_takes/448" in manifests:
    by_take = defaultdict(list)
    for bucket in s3_buckets:
        ds_base_dir = os.path.join(
            f"s3://{bucket}/{base_release_dir}", "downscaled_takes/448/"
        )
        print(ds_base_dir)
        for desc in downloader.ls(ds_base_dir, recursive=True):
            take_name = desc.path.split("downscaled_takes/")[1].split("/")[1]
            if take_name not in take_name_to_uid:
                continue
            take_uid = take_name_to_uid[take_name]
            if take_uid not in egoexo["released_takes"]:
                continue
            by_take[take_uid].append(
                PathSpecification(
                    source_path=desc.path,
                    relative_path=f"takes/{take_name}/frame_aligned_videos/downscaled/448/{desc.basename}",
                    file_type="mp4",
                    views=None,  # TODO
                    size=desc.size,
                )
            )

    for take_uid, paths in by_take.items():
        manifests["downscaled_takes/448"].append(
            ManifestEntry(
                uid=take_uid,
                paths=paths,
                benchmarks=take_uid_to_benchmarks.get(take_uid, None),
                splits=take_uid_to_splits.get(take_uid, None),
            )
        )

path = None
take_uid = None
take_name = None
by_take_and_dir = {}
for bucket in tqdm(s3_buckets):
    base_dir = os.path.join(f"s3://{bucket}/{base_release_dir}", "takes/")
    for desc in downloader.ls(base_dir, recursive=True):
        take_name = desc.path.split("takes/")[1].split("/")[0]
        if take_name not in take_name_to_uid:
            continue
        take_uid = take_name_to_uid[take_name]
        if take_uid not in egoexo["released_takes"]:
            continue
        dirname = os.path.basename(os.path.dirname(desc.path))
        key = (take_uid, dirname)
        if key not in by_take_and_dir:
            by_take_and_dir[key] = []
        by_take_and_dir[key].append(desc)


def map_take(t):
    paths = []
    take_uid = t["take_uid"]
    if take_uid not in egoexo["released_takes"]:
        return None, None

    root_dir = t["root_dir"]
    paths = []
    for streams in t["frame_aligned_videos"].values():
        for vid in streams.values():
            if vid["_s3_path"] is None:
                continue
            vid_k = (t["capture_uid"], vid.get("cam_id"))
            is_ego = capture_cam_id_to_is_ego.get(vid_k)
            views = None
            if is_ego is not None:
                views = ["ego"] if is_ego else ["exo"]

            desc = downloader.file_desc(vid["_s3_path"])
            assert desc is not None
            paths.append(
                PathSpecification(
                    source_path=vid["_s3_path"],
                    relative_path=os.path.join(root_dir, vid["relative_path"]),
                    views=views,
                    universities=[t["university_name"]],
                    file_type="mp4",
                    size=desc.size,
                )
            )

    # add vrs path (no rgb stream)
    if t["has_trimmed_vrs"]:
        for d in ["trajectory", "eye_gaze"]:
            d_paths = []
            pc_paths = []
            key = (take_uid, d)
            if key not in by_take_and_dir:
                print("Skipping", key)
                continue
            for desc in by_take_and_dir[key]:
                ext = os.path.splitext(desc.basename)[-1]
                if "semidense" in desc.basename:
                    pc_paths.append(
                        PathSpecification(
                            source_path=desc.path,
                            relative_path=os.path.join(root_dir, d, desc.basename),
                            views=None,
                            universities=[t["university_name"]],
                            file_type=ext[1:],
                            size=desc.size,
                        )
                    )
                else:
                    d_paths.append(
                        PathSpecification(
                            source_path=desc.path,
                            relative_path=os.path.join(root_dir, d, desc.basename),
                            views=None,
                            universities=[t["university_name"]],
                            file_type=ext[1:],
                            size=desc.size,
                        )
                    )

            if len(pc_paths) > 0:
                manifests["take_point_cloud"].append(
                    ManifestEntry(
                        uid=take_uid,
                        paths=pc_paths,
                        benchmarks=take_uid_to_benchmarks.get(take_uid, None),
                        splits=take_uid_to_splits.get(take_uid, None),
                    )
                )

            if d != "eye_gaze":
                assert d == "trajectory"
                manifests["take_trajectory"].append(
                    ManifestEntry(
                        uid=take_uid,
                        paths=d_paths,
                        benchmarks=take_uid_to_benchmarks.get(take_uid, None),
                        splits=take_uid_to_splits.get(take_uid, None),
                    )
                )
            else:
                assert d == "eye_gaze"
                manifests["take_eye_gaze"].append(
                    ManifestEntry(
                        uid=take_uid,
                        paths=d_paths,
                        benchmarks=take_uid_to_benchmarks.get(take_uid, None),
                        splits=take_uid_to_splits.get(take_uid, None),
                    )
                )

        sp = t["_vrs_s3_path"]
        if sp is not None:
            desc = downloader.file_desc(sp)
            assert desc is not None
            manifests["take_vrs_noimagestream"].append(
                ManifestEntry(
                    uid=take_uid,
                    paths=[
                        PathSpecification(
                            source_path=sp,
                            relative_path=os.path.join(
                                root_dir, t["vrs_relative_path"]
                            ),
                            views=["ego"],
                            universities=[t["university_name"]],
                            file_type="vrs",
                            size=desc.size,
                        )
                    ],
                    benchmarks=take_uid_to_benchmarks.get(take_uid, None),
                    splits=take_uid_to_splits.get(take_uid, None),
                )
            )

        sp_all_streams = t["_vrs_all_streams_s3_path"]
        if sp_all_streams is not None:
            desc = downloader.file_desc(sp_all_streams)
            assert desc is not None
            manifests["take_vrs"].append(
                ManifestEntry(
                    uid=take_uid,
                    paths=[
                        PathSpecification(
                            source_path=sp_all_streams,
                            relative_path=os.path.join(
                                root_dir, t["vrs_all_streams_relative_path"]
                            ),
                            views=["ego"],
                            universities=[t["university_name"]],
                            file_type="vrs",
                            size=desc.size,
                        )
                    ],
                    benchmarks=take_uid_to_benchmarks.get(take_uid, None),
                    splits=take_uid_to_splits.get(take_uid, None),
                )
            )
    return take_uid, paths


for manifest_key in ["takes", "takes_dropped"]:
    if manifest_key not in manifests:
        continue

    takes_to_map = egoexo[manifest_key]
    mapped_values = local_map(takes_to_map, map_take, 50)
    for take_uid, paths in tqdm(mapped_values):
        if take_uid is None:
            continue
        manifests[manifest_key].append(
            ManifestEntry(
                uid=take_uid,
                paths=paths,
                benchmarks=take_uid_to_benchmarks.get(take_uid, None),
                splits=take_uid_to_splits.get(take_uid, None),
            )
        )

if "take_transcription" in manifests:
    assert "take_audio" in manifests
    for take in egoexo["takes"]:
        take_uid = take["take_uid"]
        root_dir = take["root_dir"]
        audio_root_dir = os.path.join(root_dir, "audio")
        key = (take_uid, "audio")
        if key not in by_take_and_dir:
            continue

        audio_paths = []
        transcription_paths = []
        for desc in by_take_and_dir[key]:
            bn, p = desc.basename, desc.path
            size = desc.size
            if bn.endswith(".json"):
                transcription_paths += [
                    PathSpecification(
                        source_path=p,
                        relative_path=os.path.join(root_dir, "audio", bn),
                        views=None,
                        universities=None,
                        size=size,
                    )
                ]
            else:
                assert bn.endswith(".wav")
                audio_paths += [
                    PathSpecification(
                        source_path=p,
                        relative_path=os.path.join(root_dir, "audio", bn),
                        views=None,
                        universities=None,
                        size=size,
                    )
                ]

        manifests["take_audio"].append(
            ManifestEntry(
                uid=take_uid,
                paths=audio_paths,
                benchmarks=take_uid_to_benchmarks.get(take_uid, None),
                splits=take_uid_to_splits.get(take_uid, None),
            )
        )
        manifests["take_transcription"].append(
            ManifestEntry(
                uid=take_uid,
                paths=transcription_paths,
                benchmarks=take_uid_to_benchmarks.get(take_uid, None),
                splits=take_uid_to_splits.get(take_uid, None),
            )
        )

for c in tqdm(egoexo["captures"]):
    s3_bucket = c["_s3_root_dir"].split("/")[2]
    root_dir = c["root_dir"]
    traj_fp = c["_trajectory_s3_dir"]
    post_fp = c["_postsurvery_s3_path"]
    eye_gaze_fp = c["_gaze_s3_dir"]
    timesync_fp = c["_timesync_s3_path"]
    ts_dir = os.path.join(c["_s3_root_dir"], "timesync")
    traj_files = downloader.ls(traj_fp + "/", max_keys=1000) if traj_fp else []
    eye_gaze_files = (
        downloader.ls(eye_gaze_fp + "/", max_keys=1000) if eye_gaze_fp else []
    )
    ts_files = downloader.ls(ts_dir + "/", max_keys=1000)
    universities = [c["university_name"]]

    traj_paths = []
    point_cloud_paths = []
    for desc in traj_files:
        bn, s3_path = desc.basename, desc.path
        size = desc.size
        ext = os.path.splitext(bn)[-1]
        assert ext.startswith(".")
        if "semidense" in bn:
            point_cloud_paths.append(
                PathSpecification(
                    source_path=s3_path,
                    relative_path=os.path.join(root_dir, "trajectory", bn),
                    universities=universities,
                    file_type=ext[1:],
                    views=None,
                    size=size,
                )
            )
        else:
            traj_paths.append(
                PathSpecification(
                    source_path=s3_path,
                    relative_path=os.path.join(root_dir, "trajectory", bn),
                    universities=universities,
                    file_type=ext[1:],
                    views=None,
                    size=size,
                )
            )

    capture_uid = c["capture_uid"]
    manifests["capture_trajectory"].append(
        ManifestEntry(
            uid=capture_uid,
            paths=traj_paths,
            benchmarks=take_uid_to_benchmarks[capture_uid],
            splits=capture_uid_to_splits[capture_uid],
        )
    )
    manifests["capture_point_cloud"].append(
        ManifestEntry(
            uid=capture_uid,
            paths=point_cloud_paths,
            benchmarks=take_uid_to_benchmarks[capture_uid],
            splits=capture_uid_to_splits[capture_uid],
        )
    )
    eye_gaze_paths = []
    for desc in eye_gaze_files:
        ext = os.path.splitext(desc.basename)[-1]
        assert ext.startswith(".")
        eye_gaze_paths.append(
            PathSpecification(
                source_path=desc.path,
                relative_path=os.path.join(root_dir, "eye_gaze", desc.basename),
                universities=universities,
                file_type=ext[1:],
                views=None,
                size=desc.size,
            )
        )
    manifests["capture_eye_gaze"].append(
        ManifestEntry(
            uid=capture_uid,
            paths=eye_gaze_paths,
            benchmarks=take_uid_to_benchmarks[capture_uid],
            splits=capture_uid_to_splits[capture_uid],
        )
    )

    capture_paths = []
    if post_fp is not None and "public_internal" not in post_fp:
        desc = downloader.file_desc(post_fp)
        if desc is not None:
            assert desc is not None
            assert desc.basename == "post_surveys.csv"
            capture_paths.append(
                PathSpecification(
                    source_path=post_fp,
                    relative_path=os.path.join(root_dir, "post_surveys.csv"),
                    universities=universities,
                    file_type="csv",
                    views=None,
                    size=desc.size,
                )
            )

    if timesync_fp is not None:
        desc = downloader.file_desc(timesync_fp)
        assert desc is not None
        assert desc.basename == "timesync.csv"
        capture_paths.append(
            PathSpecification(
                source_path=timesync_fp,
                relative_path=os.path.join(root_dir, "timesync.csv"),
                universities=universities,
                file_type="csv",
                views=None,
                size=desc.size,
            )
        )
    for desc in ts_files:
        ext = os.path.splitext(desc.basename)[-1]
        assert ext.startswith(".")
        capture_paths.append(
            PathSpecification(
                source_path=desc.path,
                relative_path=os.path.join(root_dir, "timesync", f"{desc.basename}"),
                universities=universities,
                file_type=ext[1:],
                views=None,
                size=desc.size,
            )
        )

    manifests["captures"].append(
        ManifestEntry(
            uid=c["capture_uid"],
            paths=capture_paths,
            benchmarks=take_uid_to_benchmarks[capture_uid],
            splits=capture_uid_to_splits[capture_uid],
        )
    )


if "annotations" in manifests:
    manifests["annotations"] = []
    annotations = downloader.ls(os.path.join(release_dir, "annotations/"))
    for desc in annotations:
        bn, s3_path = desc.basename, desc.path
        if len(bn) == 0:
            continue
        if bn == "manifest.json":
            continue

        print(bn)
        benchmarks = []
        benchmark_name = "_".join(bn.split("_")[0:-1])
        benchmarks = [benchmark_name] if len(benchmark_name) > 0 else None
        if "proficiency" in benchmark_name:
            benchmarks.append("proficiency")

        manifests["annotations"].append(
            ManifestEntry(
                uid="_".join(bn.split("_")[:-1]),
                paths=[
                    PathSpecification(
                        source_path=s3_path,
                        relative_path=f"annotations/{bn}",
                        size=desc.size,
                    )
                ],
                benchmarks=benchmarks,
            )
        )

    egopose_part_subdirs = {
        "annotations": {
            "subdirs": {
                "body": ["annotation"],
                "hand": ["annotation"],
                "camera_pose": [""],
            },
            "take_uid_as_key": False,
        },
        "ego_pose_pseudo_gt": {
            "subdirs": {
                "body": ["automatic"],
                "hand": ["automatic"],
                "camera_pose": [],
            },
            "take_uid_as_key": True,
        },
    }
    if not dev_release:
        egopose_base_dir = os.path.join(release_dir, "annotations/ego_pose/")
        for split in ["train", "val", "test"]:
            for part in ["body", "hand", "camera_pose"]:
                if part in ("body", "hand") and split == "test":
                    continue

                subdir = os.path.join(egopose_base_dir, split, part)
                print(subdir)

                annotation_files = []

                # go over each potential manifest --part
                for manifest_key, metadata in egopose_part_subdirs.items():
                    for ann_dir in metadata["subdirs"][part]:
                        dst_relative_subdir = os.path.join(
                            "annotations/ego_pose/", split, part, ann_dir
                        )
                        abs_subdir = os.path.join(subdir, ann_dir)

                        if not abs_subdir.endswith("/"):
                            abs_subdir += "/"
                        ann_files = downloader.ls(abs_subdir)
                        # for bn, s3_path in ann_files:
                        for desc in ann_files:
                            bn, s3_path = desc.basename, desc.path
                            if len(bn) == 0:
                                continue
                            if bn == "manifest.json":
                                continue
                            take_uid = os.path.splitext(bn)[0]
                            uid = take_uid
                            splits = take_uid_to_splits.get(take_uid, [])
                            assert (
                                len(splits) == 1 and splits[0] == split
                            ), "split diff: {splits} vs {split}"
                            benchmarks = ["egopose", "ego_pose"]
                            if part in ["body", "hand"]:
                                benchmarks += [
                                    f"{part}_pose",
                                    f"ego_{part}_pose",
                                    f"{part}pose",
                                    f"ego{part}pose",
                                    f"ego_{part}pose",
                                ]
                            else:
                                assert part == "camera_pose"
                                for body_part in ["body", "hand"]:
                                    benchmarks += [
                                        f"{body_part}_pose",
                                        f"ego_{body_part}_pose",
                                        f"{body_part}pose",
                                        f"ego{body_part}pose",
                                        f"ego_{body_part}pose",
                                        f"camera_pose",
                                    ]

                            manifests[manifest_key].append(
                                ManifestEntry(
                                    uid=uid,
                                    benchmarks=benchmarks,
                                    splits=splits,
                                    paths=[
                                        PathSpecification(
                                            source_path=s3_path,
                                            relative_path=f"{dst_relative_subdir}/{bn}",
                                            size=desc.size,
                                        )
                                    ],
                                )
                            )

if "expert_commentary" in manifests:
    manifests["expert_commentary"] = []
    ec_base_dir = os.path.join(release_dir, "annotations/expert_commentary/")
    fs = downloader.ls(ec_base_dir, recursive=True)
    for desc in tqdm(fs):
        f = desc.path
        base_dir = f[len(ec_base_dir) :]
        take_name, expert_name = base_dir.split("/")[0:2]
        if take_name not in take_name_to_uid:
            continue
        take_uid = take_name_to_uid[take_name]
        uid = f"{take_uid}_{expert_name}"
        splits = take_uid_to_splits[take_uid]

        manifests["expert_commentary"].append(
            ManifestEntry(
                uid=uid,
                benchmarks=["expert_commentary"],
                splits=splits,
                paths=[
                    PathSpecification(
                        source_path=f,
                        relative_path=os.path.join(
                            "annotations/expert_commentary/", base_dir
                        ),
                        size=desc.size,
                    ),
                ],
            )
        )

if "features/omnivore_video" in manifests:
    for feature_name in ["omnivore_video", "maws_clip_2b"]:
        feature_files = downloader.ls(
            os.path.join(release_dir, "features/", feature_name + "/")
        )
        by_take = defaultdict(list)
        for desc in feature_files:
            file_name, path = desc.basename, desc.path
            if file_name.endswith("yaml"):
                manifests[f"features/{feature_name}"].append(
                    ManifestEntry(
                        uid="omnivore_video_config",
                        paths=[
                            PathSpecification(
                                source_path=path,
                                relative_path=f"features/{feature_name}/{file_name}",
                                size=desc.size,
                            )
                        ],
                    )
                )
                continue

            if file_name == "" or file_name == "manifest.json":
                continue
            take_uid, cam_id, stream_id = file_name.split("_")
            if take_uid not in egoexo["released_takes"]:
                continue

            t = take_by_take_uid[take_uid]
            vid_k = (t["capture_uid"], cam_id)
            is_ego = capture_cam_id_to_is_ego.get(vid_k)
            views = None
            if is_ego is not None:
                views = ["ego"] if is_ego else ["exo"]
            by_take[take_uid].append(
                PathSpecification(
                    source_path=path,
                    relative_path=f"features/{feature_name}/{file_name}",
                    views=views,
                    universities=[t["university_name"]],
                    file_type="pt",
                    size=desc.size,
                )
            )

        for take_uid, paths in by_take.items():
            manifests[f"features/{feature_name}"].append(
                ManifestEntry(
                    uid=take_uid,
                    paths=paths,
                    benchmarks=take_uid_to_benchmarks.get(take_uid, None),
                    splits=take_uid_to_splits.get(take_uid, None),
                )
            )


all_bs = set()
for k, v in manifests.items():
    out_dir = os.path.join(release_dir, f"{k}/")
    manifest_file = os.path.join(out_dir, "manifest.json")
    for m in v:
        for b in m.benchmarks or []:
            all_bs.add(b)

    print(k, len(v))
    print(f"will output to: {out_dir}")
    print(f"{manifest_file}")
    pathmgr.mkdirs(out_dir, exist_ok=True)
    pathmgr.rm(manifest_file)
    with pathmgr.open(manifest_file, "w") as out_f:
        json_data = manifest_dumps(v)
        out_f.write(json_data)


def print_manifest_stats(release_dir: str):
    manifests = [
        "metadata",
        "annotations",
        "takes",
        "captures",
        "take_trajectory",
        "take_eye_gaze",
        "take_point_cloud",
        "take_vrs",
        "take_vrs_noimagestream",
        "capture_trajectory",
        "capture_eye_gaze",
        "capture_point_cloud",
        "downscaled_takes/448",
        "features/omnivore_video",
    ]
    if "v2" in release_dir:
        manifests += [
            "features/maws_clip_2b",
            "ego_pose_pseudo_gt",
            "expert_commentary",
            "take_transcription",
            "take_audio",
        ]

    manifest_descs = {
        "metadata": "See [metadata](/data/metadata)",
        "annotations": "All the [annotations](/annotations/) in Ego-Exo4D",
        "takes": "Frame aligned video files associated to the [takes](/data/takes) ",
        "captures": "Timesync and post-survey data at the capture level (multiple takes) ",
        "take_trajectory": "[Trajectories](/data/mps/#trajectory) trimmed at each take",
        "take_eye_gaze": "[Eye gaze](/data/mps/#eye-gaze) for each take (3D & 2D)",
        "take_vrs_noimagestream": "VRS files for each take without image stream data (video data within MP4 containers with `--parts takes`)",
        "take_vrs": "VRS files for each take",
        "take_point_cloud": "[Point clouds](/data/mps#point-clouds) for each take",
        "take_transcription": "[Audio transcriptions](https://github.com/facebookresearch/Ego4d/blob/main/ego4d/egoexo/scripts/extract_audio_transcribe.py#L22-L47) for each take",
        "take_audio": "[Audio files](https://github.com/facebookresearch/Ego4d/blob/main/ego4d/egoexo/scripts/extract_audio_transcribe.py#L22-L47) for the egocentric aria camera ",
        "capture_trajectory": "[Trajectory](/data/mps#trajectory) at the capture-level",
        "capture_eye_gaze": "[Eye gaze](/data/mps#eye_gaze) at the capture-level (3D) ",
        "capture_point_cloud": "[Point clouds](data/mps/#point-clouds) for each capture",
        "downscaled_takes/448": "[Downscaled takes](data/downscaled_takes/) at 448px on the shortest side",
        "features/omnivore_video": "Omnivore video [features](/data/features)",
        "features/maws_clip_2b": "[MAWS CLIP](https://github.com/facebookresearch/maws) ([ViT-2b](https://github.com/facebookresearch/maws?tab=readme-ov-file#maws-pretrained-models)) [features](/data/features) for each frame of video",
        "ego_pose_pseudo_gt": "Pseudo-ground truth data for [Ego Pose](/annotations/ego_pose/)",
        "expert_commentary": "[Commentaries](/annotations/expert_commentary) for each expert (audio recordings)",
        "all": "All data within the release (you can use `--parts all`) ",
        "default": "The default set of data in the release (you can use `--parts default` or provide no parts)",
    }

    manifests_by_name = {}
    for m_name in tqdm(manifests):
        m_path = os.path.join(release_dir, m_name, "manifest.json")
        m = manifest_loads(pathmgr.open(m_path).read())
        manifests_by_name[m_name] = m

    files_by_mname = {}
    for m_name, ms in tqdm(manifests_by_name.items()):
        curr_files = {}
        for m in ms:
            for path in m.paths:
                assert path.size is not None
                curr_files[path.relative_path] = path.size
        files_by_mname[m_name] = curr_files

    file_size_by_mname = {k: sum(v.values()) for k, v in files_by_mname.items()}

    default = [
        "metadata",
        "captures",
        "takes",
        "take_trajectory",
        "take_vrs_noimagestream",
        "annotations",
    ]

    default_size = 0
    all_file_size = 0

    for k, v in file_size_by_mname.items():
        assert k in manifest_descs, k
        if k in default:
            default_size += v
        all_file_size += v
    file_size_by_mname["all"] = all_file_size
    file_size_by_mname["default"] = default_size

    print("| Part | Size (GB) | Description |")
    print("| -----|-----------|-------------|")
    for k, v in file_size_by_mname.items():
        assert k in manifest_descs
        size = v / 10**9
        if k in default:
            print(f"| **{k}** | {size:.3f} | {manifest_descs[k]} |")
        else:
            print(f"| {k} | {size:.3f} | {manifest_descs[k]} |")


print_manifest_stats(release_dir)



from dataclasses import dataclass, field
from typing import List, Optional

from dataclasses_json import dataclass_json


@dataclass_json
@dataclass(frozen=True)
class PathSpecification:
    source_path: str
    relative_path: str
    views: Optional[List[str]] = field(
        default_factory=lambda: None, compare=False, hash=False
    )
    universities: Optional[List[str]] = field(
        default_factory=lambda: None, compare=False, hash=False
    )
    file_type: Optional[str] = field(
        default_factory=lambda: None, compare=False, hash=False
    )
    size: Optional[int] = field(default_factory=lambda: None, compare=False, hash=False)
    checksum: Optional[str] = field(
        default_factory=lambda: None, compare=False, hash=False
    )


@dataclass_json
@dataclass(frozen=True)
class ManifestEntry:
    uid: str
    paths: List[PathSpecification]
    splits: Optional[List[str]] = field(
        default_factory=lambda: None, compare=False, hash=False
    )
    benchmarks: Optional[List[str]] = field(
        default_factory=lambda: None, compare=False, hash=False
    )


def manifest_dumps(xs: List[ManifestEntry]) -> str:
    return ManifestEntry.schema().dumps(xs, many=True)  # pyre-ignore


def manifest_loads(data: str) -> List[ManifestEntry]:
    return ManifestEntry.schema().loads(data, many=True)  # pyre-ignore



import argparse
import os
import sys
import threading
import traceback
from concurrent.futures import ThreadPoolExecutor
from typing import Callable, List, Optional, Tuple, TypeVar

from ego4d.cli.progressbar import DownloadProgressBar
from ego4d.cli.universities import UNIV_TO_BUCKET

from ego4d.internal.download.manifest import (
    manifest_loads,
    ManifestEntry,
    PathSpecification,
)
from ego4d.internal.s3 import S3Downloader
from iopath.common.file_io import PathManager
from iopath.common.s3 import S3PathHandler
from tqdm.auto import tqdm

T = TypeVar("T")
U = TypeVar("U")


# TODO: remove iopath dependency
def _create_pathmgr(s3_profile: Optional[str]):
    pathmgr = PathManager()
    pathmgr.register_handler(S3PathHandler(profile=s3_profile))
    return pathmgr


def _s3_path_join(a: str, b: str) -> str:
    assert a.startswith("s3://")
    return os.path.join(a, b).replace("\\", "/")  # hacky, but it works


def _path_ok(p: PathSpecification, args) -> bool:
    ok = True
    if (
        args.views is not None
        and p.views is not None
        and len(set(p.views) & args.views) == 0
    ):
        ok = False
    if (
        args.universities is not None
        and p.universities is not None
        and len(set(p.universities) & args.universities) == 0
    ):
        ok = False
    return ok


def _manifest_ok(m: ManifestEntry, args) -> bool:
    ok = True
    if (
        args.splits is not None
        and m.splits is not None
        and len(set(m.splits) & args.splits) == 0
    ):
        ok = False
    if (
        args.benchmarks is not None
        and m.benchmarks is not None
        and len(set(m.benchmarks) & args.benchmarks) == 0
    ):
        ok = False
    if args.uids is not None and m.uid is not None and len({m.uid} & args.uids) == 0:
        ok = False
    return ok


def map_all(
    values: List[T],
    map_fn: Callable[[T, Optional[S3Downloader]], Tuple[U, T, Optional[str]]],
    num_workers: int,
    s3_profile: Optional[str],
    needs_downloader: bool,
    progress_on_bytes: bool,
    total_bytes: Optional[int],
) -> Tuple[List[Tuple[T, U]], List[Tuple[T, str]]]:
    # NOTE: boto3 is not thread safe
    thread_data = threading.local()
    callback = None

    def initializer():
        if needs_downloader:
            thread_data.downloader = S3Downloader(
                s3_profile,
                num_workers=num_workers,
                callback=callback,
            )

    def wrap_map(x):
        kwargs = {}
        if needs_downloader:
            kwargs["downloader"] = thread_data.downloader

        return map_fn(x, **kwargs)  # pyre-ignore

    failures = []
    ret = []
    iterator_wrapper = tqdm
    if progress_on_bytes:
        progress = DownloadProgressBar(total_bytes)
        iterator_wrapper = lambda x, total: x
        callback = progress.update if progress is not None else None

    with ThreadPoolExecutor(max_workers=num_workers, initializer=initializer) as pool:
        for value, key, err in iterator_wrapper(
            pool.map(wrap_map, values), total=len(values)
        ):
            if value is None:
                failures.append((key, err))
            else:
                ret.append((key, value))
    return ret, failures


def main(args):
    base_dir = args.base_dir
    release_name = args.release
    parts = set(args.parts)
    out_dir = args.out_dir
    num_workers = args.num_workers
    s3_profile = args.s3_profile
    yes = args.yes
    delete_yes = args.delete_yes
    args.uids = set(args.uids) if args.uids is not None else None
    args.benchmarks = set(args.benchmarks) if args.benchmarks is not None else None
    args.splits = set(args.splits) if args.splits is not None else None
    args.views = set(args.views) if args.views is not None else None
    args.universities = (
        set(args.universities) if args.universities is not None else None
    )

    if "all" in parts:
        if len(parts) > 1:
            print(
                "WARN: you have supplied all parts; ignoring the other parts supplied"
            )

        parts = {
            "metadata",
            "annotations",
            "takes",
            "take_vrs_noimagestream",
            "take_trajectory",
            "take_eye_gaze",
            "take_point_cloud",
            "take_vrs",
            "take_transcription",
            "take_audio",
            "captures",
            "capture_trajectory",
            "capture_eye_gaze",
            "capture_point_cloud",
            "downscaled_takes/448",
            "features/omnivore_video",
            "features/maws_clip_2b",
            "ego_pose_pseudo_gt",
            "expert_commentary",
        }
        print(f"NOTE: all part supplied; equivalent to: --parts {' '.join(parts)}")
    if "default" in parts:
        parts_to_add = {
            "metadata",
            "captures",
            "takes",
            "take_trajectory",
            "take_vrs_noimagestream",
            "annotations",
        }
        print(
            f"NOTE: default part supplied; equivalent to: --parts {' '.join(parts_to_add)}"
        )
        parts.remove("default")
        parts.update(parts_to_add)

    assert (
        release_name is not None
    ), """You must provide a release name via `--release`.
If you meant to download the public release, please use the script `ego4d/egoexo/download/cli.py` instead.
"""

    def check_file(
        path_expected_size: Tuple[PathSpecification, int],
    ) -> Tuple[Optional[int], Tuple[PathSpecification, int], Optional[str]]:
        path, expected_size = path_expected_size
        out_path = os.path.join(out_dir, path.relative_path)
        try:
            if not os.path.exists(out_path):
                return 0, path_expected_size, None
            size = os.path.getsize(out_path)
            ok = size == expected_size
            return size if ok else 0, path_expected_size, None
        except Exception:
            return None, path_expected_size, traceback.format_exc()

    def get_size(
        path: PathSpecification,
        downloader: Optional[S3Downloader],
    ) -> Tuple[Optional[int], PathSpecification, Optional[str]]:
        assert downloader is not None
        try:
            if not path.source_path.startswith("s3://"):
                return None, path, "path is not an S3 path"
            return downloader.obj(path.source_path).content_length, path, None
        except Exception:
            return None, path, traceback.format_exc()

    def download(
        path_size_pair: Tuple[PathSpecification, int],
        downloader: Optional[S3Downloader],
    ) -> Tuple[Optional[int], Tuple[PathSpecification, int], Optional[str]]:
        assert downloader is not None
        path, _ = path_size_pair
        try:
            out_path = os.path.join(out_dir, path.relative_path)
            downloader.copy(path.source_path, out_path)
            size, _, err = check_file(path_size_pair)
            return size, path_size_pair, err
        except Exception:
            return None, path_size_pair, traceback.format_exc()

    os.makedirs(out_dir, exist_ok=True)
    assert os.path.isdir(out_dir), f"output dir {out_dir} is not a directory"

    release_dir = _s3_path_join(base_dir, release_name)
    if not release_dir.endswith("/"):
        release_dir += "/"

    pathmgr = _create_pathmgr(s3_profile)

    num_paths = 0
    all_paths = []
    part_errs = []
    print("Obtaining part metadata ...", flush=True)
    valid_benchmark_names = set()
    for part in tqdm(parts):
        manifest_path = _s3_path_join(_s3_path_join(release_dir, part), "manifest.json")
        if not pathmgr.exists(manifest_path):
            part_errs.append(part)
            continue

        ms = manifest_loads(pathmgr.open(manifest_path).read())
        valid_benchmark_names.update({b for m in ms for b in m.benchmarks or []})
        for m in ms:
            num_paths += len(m.paths)
            if not _manifest_ok(m, args):
                continue

            all_paths.extend([p for p in m.paths if _path_ok(p, args)])

    print("Done", flush=True)

    invalid_benchmarks = (
        args.benchmarks - valid_benchmark_names if args.benchmarks else set()
    )
    if len(invalid_benchmarks) > 0:
        print(f"WARN: Provided invalid benchmarks: {invalid_benchmarks} for parts provided (parts may not be associated to benchmarks)")
        if len(valid_benchmark_names) == 0:
            print("Parts are not associated to benchmarks")
        else:
            print("Valid benchmark names include: ")
            print()
            for b in valid_benchmark_names:
                print(b)
            print()
        print("Proceeding anyway via setting benchmarks to the empty set ...")
        args.benchmarks = set()

    if len(part_errs) == len(parts):
        print("ERROR: could not get manifests for all parts (nothing to download)")
        print("Did you supply the right parts?")
        print(
            """NOTE: Did you configure your AWS client correctly? See: https://github.com/facebookresearch/Ego4d/issues/277

This could also be due to your internet connection.

If you are located in China, please try using a VPN. Please refer to these posts:
- https://github.com/facebookresearch/Ego4d/issues/223
- https://github.com/facebookresearch/Ego4d/issues/162
- https://github.com/facebookresearch/Ego4d/issues/284
"""
        )
        print("Quitting ...")
        sys.exit(1)
    elif len(part_errs) > 0:
        print(f"ERROR: some supplied parts ({part_errs}) do not exist")
        print("NOTE: did you spell the parts correctly?")
        print("Quitting ...")
        sys.exit(2)

    if num_paths != len(all_paths):
        print(f"Filtered {num_paths} -> {len(all_paths)} files")

    path_size_pairs = [(x, x.size) for x in all_paths if x.size is not None]
    not_cached_paths = [x for x in all_paths if x.size is None]

    if len(not_cached_paths) > 0:
        print("Determining what to download ...")
        extra_path_size_pairs, s3_stat_failures = map_all(
            not_cached_paths,
            map_fn=get_size,
            num_workers=num_workers,
            s3_profile=s3_profile,
            needs_downloader=True,
            progress_on_bytes=False,
            total_bytes=None,
        )
        s3_zero_sizes = [path for path, size in extra_path_size_pairs if size == 0]
        path_size_pairs += [
            (path, size) for path, size in extra_path_size_pairs if size != 0
        ]
        all_s3_stat_failures = len(s3_zero_sizes) + len(s3_stat_failures)
        if all_s3_stat_failures > 0:
            bucket_failures = {
                path.source_path.split("s3://")[1].split("/")[0]
                for path, _ in s3_stat_failures
            }
            print(
                f"""WARN: failed to get stats for {all_s3_stat_failures} files, will skip. [zero_size={len(s3_zero_sizes)}, exceptions={all_s3_stat_failures}.

    This could be due to your internet connection.

    If you are located in China, please try using a VPN. Please refer to these posts:
    - https://github.com/facebookresearch/Ego4d/issues/223
    - https://github.com/facebookresearch/Ego4d/issues/162
    - https://github.com/facebookresearch/Ego4d/issues/284

    Additional debugging information:
    - S3 Buckets failed: {bucket_failures}
    """
            )

    total_size_bytes = sum(x for _, x in path_size_pairs if x is not None)
    total_size_gib = total_size_bytes / 1024**3

    print("Checking current download status ...")
    curr_paths, curr_stats_failures = map_all(
        path_size_pairs,
        map_fn=check_file,  # pyre-ignore
        num_workers=num_workers,
        s3_profile=s3_profile,
        needs_downloader=False,
        progress_on_bytes=False,
        total_bytes=None,
    )
    # TODO: refactor such that we can tell the number of files to be updated
    existing_paths = {
        path_size_pair
        for path_size_pair, actual_size in curr_paths
        if actual_size is not None and actual_size > 0
    }
    existing_gib = sum(size / 1024**3 for _, size in existing_paths if size is not None)
    existing_len = len(existing_paths)
    progress_percent = existing_gib / total_size_gib if total_size_gib > 0 else 1.0

    print(
        f"Downloaded: {progress_percent:.3%} = {existing_gib:.3f}GiB / {total_size_gib:.3f}GiB ({existing_len} / {len(path_size_pairs)} files) downloaded"
    )
    if len(curr_stats_failures) > 0:
        print(f"WARN: {len(curr_stats_failures)} failed to stat")

    ps_to_dl = {
        path: size
        for path, size in path_size_pairs
        if (path, size) not in existing_paths and size > 0
    }
    if len(ps_to_dl) == 0 and not args.force:
        print("Everything has been downloaded. Bye.")
        sys.exit(0)

    if args.force:
        print("Forcing everything to be re-downloaded ...")
        ps_to_dl = {path: size for path, size in path_size_pairs}

    assert all(x is not None for x in ps_to_dl.values())
    expected_gb = sum(x / 1024**3 for x in ps_to_dl.values() if x is not None)
    confirm = False
    if not yes:
        response = input(
            f"Expected size of downloaded files is "
            f"{expected_gb:.1f} GiB ({len(ps_to_dl)} files). "
            f"Do you want to start the download? ([y]/n): "
        )
        if response.lower() in ["yes", "y", ""]:
            confirm = True
        else:
            confirm = False
    else:
        confirm = True

    if not confirm:
        print("Aborting...")
        sys.exit(0)

    if args.delete:
        print("Scanning for files to delete ...")
        files_that_exist = []
        for dirpath, _, filenames in os.walk(out_dir):
            files_that_exist.extend([os.path.join(dirpath, f) for f in filenames])

        files_that_exist = set(files_that_exist)
        files_to_delete = files_that_exist - {
            os.path.join(out_dir, x[0].relative_path) for x in path_size_pairs
        }
        if len(files_to_delete) > 0:
            response = input(
                f"Will delete: {len(files_to_delete)} files (there is {len(files_that_exist)} total files) "
                f"Continue? (y/[n]): "
            )
            if not delete_yes:
                if response.lower() in ["yes", "y"]:
                    confirm = True
                else:
                    confirm = False
            else:
                confirm = True

            if not confirm:
                print("Aborting...")
                sys.exit(0)

            for f in tqdm(files_to_delete):
                os.remove(f)

    print("Preparing output directories ...")
    all_out_dirs = {
        os.path.join(out_dir, os.path.dirname(x.relative_path)) for x in all_paths
    }
    for x in tqdm(all_out_dirs):
        os.makedirs(x, exist_ok=True)

    print("Downloading ...")
    assert all(size is not None for size in ps_to_dl.values())
    paths_to_fetch = [
        (path, size) for path, size in ps_to_dl.items() if size is not None
    ]
    total_bytes_to_fetch = sum(size for _, size in paths_to_fetch)
    print(f"Fetching: {total_bytes_to_fetch/1024**3:.3f}GiB")
    paths_downloaded, dl_failures = map_all(
        paths_to_fetch,  # pyre-ignore
        map_fn=download,  # pyre-ignore
        num_workers=num_workers,
        s3_profile=s3_profile,
        needs_downloader=True,
        progress_on_bytes=True,
        total_bytes=total_bytes_to_fetch,
    )

    integrity_errs = [path for path, size in paths_downloaded if size == 0]
    if len(dl_failures) > 0:
        num_all_failures = len(dl_failures) + len(integrity_errs)
        print(
            f"WARN: failed to fetch {num_all_failures} files [integrity={len(integrity_errs)}, exceptions={len(dl_failures)}]"
        )
        print("Please retry the download (... returning with error code 2)")
        sys.exit(2)
    print(flush=True)


def create_arg_parse(script_name: str, base_dir: str, release_name: Optional[str]):
    parser = argparse.ArgumentParser(
        usage=f"""
    EgoExo downloader CLI

    Simple usage:
        {script_name} -o <out_dir>

    Advanced usage examples:
        - Download point clouds and annotations 
            {script_name} -o <out_dir> --parts annotations take_point_cloud -y
        - Download VRS files (with RGB stream) for a take
            {script_name} -o <out_dir> --parts take_vrs --uids <uid1>

"""
    )
    parser.add_argument(
        "-o",
        "--out_dir",
        type=str,
        default=None,
        help="Which folder to download the data to",
        required=True,
    )
    parser.add_argument(
        "--parts",
        type=str,
        nargs="+",
        default=["default"],
        help="""
What parts of the dataset to download, one of:
- metadata
- annotations
- takes
- take_vrs_noimagestream
- take_trajectory
- take_eye_gaze
- take_point_cloud
- take_vrs
- take_audio
- take_transcription
- captures
- capture_trajectory
- capture_eye_gaze
- capture_point_cloud
- downscaled_takes/448
- features/omnivore_video
- features/maws_clip_2b
- ego_pose_pseudo_gt
- expert_commentary

By default the following parts will be downloaded: {metadata, captures, takes, trajectory, annotations}.

Example usage: --parts annotations point_cloud eye_gaze
""",
    )
    parser.add_argument(
        "--uids",
        type=str,
        nargs="+",
        default=None,
        help="what uids to filter for takes or captures",
    )
    parser.add_argument(
        "--benchmarks",
        type=str,
        nargs="+",
        default=None,
        help="""
Data relating to particular benchmarks will be only included. Valid benchmark
values are: keystep, proficiency, proficiency_demonstration, proficiency_demonstrator, relations, translation, correspondence, atomic_action_descriptions, egopose
""",
    )
    parser.add_argument(
        "--splits",
        type=str,
        nargs="+",
        default=None,
        help="""
Data relating to the train/val/test dataset splits. Valid values are {train, val, test}.
""",
    )
    parser.add_argument(
        "--views",
        type=str,
        nargs="+",
        default=None,
        help="""
Data relating to a particular view. Valid values are {ego, exo}.
""",
    )
    parser.add_argument(
        "-u",
        "--universities",
        type=str,
        nargs="+",
        default=None,
        help=f"""
Data relating to a particular university. Valid values are:
{", ".join(UNIV_TO_BUCKET.keys())}
""",
    )
    parser.add_argument(
        "--num_workers",
        type=int,
        default=15,
        help="number of workers to perform download ops",
    )
    parser.add_argument(
        "-y",
        "--yes",
        default=False,
        help="don't prompt to confirm",
        action="store_true",
    )
    parser.add_argument(
        "-dy",
        "--delete_yes",
        default=False,
        help="don't prompt to confirm deletion",
        action="store_true",
    )
    parser.add_argument(
        "-d",
        "--delete",
        default=False,
        help="delete unused files in the directory",
        action="store_true",
    )
    parser.add_argument(
        "--force",
        default=False,
        help="force a download of all files",
        action="store_true",
    )
    parser.add_argument(
        "--release",
        type=str,
        default=release_name,
        help="name/version of the release",
    )
    parser.add_argument(
        "--s3_profile",
        type=str,
        default="default",
        help="profile to use for S3",
    )
    parser.add_argument(
        "--base_dir",
        type=str,
        default=base_dir,
        help="base directory for download (ADVANCED usage, do not change unless you know what you're doing)",
    )
    return parser


def internal_main():
    parser = create_arg_parse(
        script_name="egoexo_internal",
        release_name=None,
        base_dir="s3://ego4d-consortium-sharing/egoexo/releases/",
    )
    args = parser.parse_args()
    main(args)


if __name__ == "__main__":
    internal_main()



from ego4d.internal.download.cli import create_arg_parse, main as download_main


def main() -> None:
    parser = create_arg_parse(
        script_name="egoexo",
        release_name="v2",
        base_dir="s3://ego4d-consortium-sharing/egoexo-public/",
    )
    args = parser.parse_args()
    download_main(args)


if __name__ == "__main__":
    main()



# Copyright (c) Meta Platforms, Inc. and affiliates. All Rights Reserved.

import json
import os
import random
from dataclasses import dataclass
from typing import Optional

import av
import hydra
import pandas as pd
import torch
import torchvision.transforms as T
from ego4d.features.config import FeatureExtractConfig, load_model, Video
from ego4d.features.extract_features import extract_features
from PIL import Image


@dataclass
class RunInferenceConfig(FeatureExtractConfig):
    dataset_type: str = "k400"
    dataset_dir: str = "/datasets01/Kinetics400_Frames/videos/"
    set_to_use: str = "val"
    num_examples: int = 5
    top_k: int = 2
    seed: Optional[int] = None


def _video_info(path: str):
    with av.open(path) as container:
        if len(container.streams.video) == 0:
            return {
                "num_frames": None,
                "fps": None,
            }
        return {
            "num_frames": container.streams.video[0].frames,
            "fps": container.streams.video[0].average_rate,
        }


def _load_kinetics_class_names():
    # wget \
    # https://dl.fbaipublicfiles.com/pyslowfast/dataset/class_names/kinetics_classnames.json
    with open(
        "/private/home/miguelmartin/ego4d/ego4d_public/kinetics_classnames.json", "r"
    ) as f:
        kinetics_classnames = json.load(f)

    # Create an id to label name mapping
    kinetics_id_to_classname = {}
    for k, v in kinetics_classnames.items():
        kinetics_id_to_classname[v] = str(k).replace('"', "")
    return kinetics_id_to_classname


def _load_imagenet_label_dir_dict(dataset_dir: str):
    path = os.path.join(dataset_dir, "labels.txt")
    df = pd.read_csv(path, names=["dir_name", "class_name"], header=None)
    return dict(zip(df.dir_name.tolist(), df.class_name.tolist()))


def _load_imagenet_class_names():
    # NOTE
    # this class mapping is for omnivore
    # this may or may not be applicable to you
    # wget https://s3.amazonaws.com/deep-learning-models/image-models/imagenet_class_index.json
    with open(
        "/private/home/miguelmartin/ego4d/ego4d_public/imagenet_class_index.json", "r"
    ) as f:
        imagenet_classnames = json.load(f)

    # Create an id to label name mapping
    imagenet_id_to_classname = {}
    for k, v in imagenet_classnames.items():
        imagenet_id_to_classname[k] = v[1]
    return imagenet_id_to_classname


def inference_imagenet(config: RunInferenceConfig):
    # taken from https://github.com/facebookresearch/omnivore/blob/main/inference_tutorial.ipynb
    label_mapping = _load_imagenet_class_names()
    label_dir_dict = _load_imagenet_label_dir_dict(config.dataset_dir)

    image_set_dir = os.path.join(config.dataset_dir, config.set_to_use)
    labels = os.listdir(image_set_dir)

    for _ in range(config.num_examples):
        expected_label = random.sample(labels, 1)[0]
        expected_label_value = label_dir_dict[expected_label]

        images_dir = os.path.join(image_set_dir, expected_label)
        images_in_dir = os.listdir(images_dir)

        image_path = random.sample(images_in_dir, 1)[0]
        image_path = f"{images_dir}/{image_path}"

        image = Image.open(image_path).convert("RGB")
        image_transform = T.Compose(
            [
                T.Resize(224),
                T.CenterCrop(224),
                T.ToTensor(),
                T.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
            ]
        )
        image = image_transform(image)
        image = image[None, :, None, ...].to(config.inference_config.device)
        model = load_model(config, patch_final_layer=False)
        predictions = model(image)

        pred_classes = torch.nn.Softmax(dim=1)(predictions).squeeze()
        top_k = pred_classes.topk(k=config.top_k)
        predictions_strs = [
            f"- {label_mapping[str(idx.item())]}: {prob.item():.2%}"
            for prob, idx in zip(top_k.values, top_k.indices)
        ]
        predictions_summary = "\n".join(predictions_strs)
        print(
            f"""Example: {image_path}

Expected label: {expected_label_value}
Predicted labels:
{predictions_summary}
    """
        )


def inference_k400(config: RunInferenceConfig):
    """
    Follows https://pytorchvideo.org/docs/tutorial_torchhub_inference

    Assumes kinetics is structured as:
    - <basedir>/train
    - <basedir>/val
    - <basedir>/test
    """
    video_set_dir = os.path.join(config.dataset_dir, config.set_to_use)

    model = load_model(config, patch_final_layer=False)
    label_mapping = _load_kinetics_class_names()
    labels = os.listdir(video_set_dir)

    for i in range(config.num_examples):
        expected_label = random.sample(labels, 1)[0]

        videos_dir = os.path.join(video_set_dir, expected_label)
        videos_in_dir = [x for x in os.listdir(videos_dir) if x.endswith("mp4")]
        video_path = random.sample(videos_in_dir, 1)[0]  # pick an example
        video_path = f"{videos_dir}/{video_path}"

        input_id = "logit"
        info = _video_info(video_path)
        videos = [Video(input_id, video_path, info["num_frames"])]

        predictions = extract_features(
            videos=videos,
            config=config,
            model=model,
            log_info=False,
            silent=True,
            assert_feature_size=False,
        ).result[input_id]

        pred_classes = torch.nn.Softmax(dim=0)(predictions.mean(0))
        top_k = pred_classes.topk(k=config.top_k)
        predictions_strs = [
            f"- {label_mapping[idx.item()]}: {prob.item():.2%}"
            for prob, idx in zip(top_k.values, top_k.indices)
        ]
        predictions_summary = "\n".join(predictions_strs)

        print(
            f"""Example {i+1}: {video_path}

Expected label: {expected_label}
Predicted labels:
{predictions_summary}
"""
        )


@hydra.main(config_path="configs", config_name=None)
def main(config: RunInferenceConfig):
    if config.seed is not None:
        random.seed(config.seed)

    assert config.dataset_type in ("k400", "imagenet")
    if config.dataset_type == "k400":
        inference_k400(config)
    else:
        inference_imagenet(config)


if __name__ == "__main__":
    main()  # pyre-ignore



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

import threading
from typing import Optional

import tqdm


class DownloadProgressBar:
    """
    Thread-safe progress bar for tracking downloads based on bytes.
    """

    def __init__(self, total_size_bytes: Optional[int]):
        self.__tqdm = tqdm.tqdm(
            total=total_size_bytes if total_size_bytes else None,
            unit="iB",
            unit_scale=True,
            unit_divisor=1024,
        )
        self.__lock = threading.Lock()

    def update(self, num_bytes: int) -> None:
        with self.__lock:
            self.__tqdm.update(n=num_bytes)



import math
from dataclasses import dataclass
from typing import Any, List

import submitit


@dataclass
class SlurmConfig:
    slurm_log_folder: str
    timeout_min: int
    constraint: str
    slurm_partition: str
    slurm_array_parallelism: int
    gpus_per_node: int
    cpus_per_task: int
    run_locally: bool = False


def batch_it(things: List[Any], batch_size: int) -> List[List[Any]]:
    num_batches: int = math.ceil(len(things) / batch_size)

    result = []
    for i in range(num_batches):
        result.append(things[i * batch_size : (i + 1) * batch_size])
    return result


def create_executor(config: SlurmConfig, num_batches: int):
    if config.run_locally:
        executor = submitit.LocalExecutor(folder=config.slurm_log_folder)
    else:
        executor = submitit.AutoExecutor(folder=config.slurm_log_folder)

    executor.update_parameters(
        timeout_min=config.timeout_min,
        slurm_constraint=config.constraint,
        slurm_partition=config.slurm_partition,
        slurm_array_parallelism=min(config.slurm_array_parallelism, num_batches),
        gpus_per_node=config.gpus_per_node,
        cpus_per_task=config.cpus_per_task,
    )
    return executor



from dataclasses import dataclass

import numpy as np

from ego4d.research.util.lzstring import decompress_from_encoded_uri
from pycocotools import mask as mask_utils


@dataclass
class Mask:
    width: int
    height: int
    encoded_mask: str


def decode_mask(mask: dict) -> np.ndarray:
    w = mask["width"]
    h = mask["height"]
    encoded_mask = mask["encodedMask"]
    return decode_mask_obj(Mask(width=w, height=h, encoded_mask=encoded_mask))


def decode_mask_obj(mask: Mask) -> np.ndarray:
    decomp_string = decompress_from_encoded_uri(mask.encoded_mask)
    decomp_encoded = decomp_string.encode()  # pyre-ignore
    rle_obj = {
        "size": [mask.height, mask.width],
        "counts": decomp_encoded,
    }

    output = mask_utils.decode(rle_obj)
    return output


def blend_mask(
    input_img: np.ndarray, binary_mask: np.ndarray, alpha: float = 0.5
) -> np.ndarray:
    if input_img.ndim == 2:
        return input_img

    mask_image = np.zeros(input_img.shape, np.uint8)
    mask_image[:, :, 1] = 255
    mask_image = mask_image * np.repeat(binary_mask[:, :, np.newaxis], 3, axis=2)

    blend_image = input_img[:, :, :]
    pos_idx = binary_mask > 0
    for ind in range(input_img.ndim):
        ch_img1 = input_img[:, :, ind]
        ch_img2 = mask_image[:, :, ind]
        ch_img3 = blend_image[:, :, ind]
        ch_img3[pos_idx] = alpha * ch_img1[pos_idx] + (1 - alpha) * ch_img2[pos_idx]
        blend_image[:, :, ind] = ch_img3
    return blend_image



import numpy as np


# Taken from: https://raw.githubusercontent.com/wykang/Charades/master/utils/map.py
def mAP(submission_array, gt_array):
    """Returns mAP, weighted mAP, and AP array"""
    m_aps = []
    n_classes = submission_array.shape[1]
    for oc_i in range(n_classes):
        sorted_idxs = np.argsort(-submission_array[:, oc_i])
        tp = gt_array[:, oc_i][sorted_idxs] == 1
        fp = np.invert(tp)
        n_pos = tp.sum()
        if n_pos < 0.1:
            m_aps.append(float("nan"))
            continue
        fp.sum()
        f_pcs = np.cumsum(fp)
        t_pcs = np.cumsum(tp)
        prec = t_pcs / (f_pcs + t_pcs).astype(float)
        avg_prec = 0
        for i in range(submission_array.shape[0]):
            if tp[i]:
                avg_prec += prec[i]
        m_aps.append(avg_prec / n_pos.astype(float))
    m_aps = np.array(m_aps)
    m_ap = np.mean(m_aps)
    w_ap = m_aps * gt_array.sum(axis=0) / gt_array.sum().sum().astype(float)
    return m_ap, w_ap, m_aps


def nanmap(submission_array, gt_array):
    """Returns mAP, weighted mAP, and AP array"""
    m_aps = []
    n_classes = submission_array.shape[1]
    for oc_i in range(n_classes):
        sorted_idxs = np.argsort(-submission_array[:, oc_i])
        tp = gt_array[:, oc_i][sorted_idxs] == 1
        fp = np.invert(tp)
        n_pos = tp.sum()
        if n_pos < 0.1:
            m_aps.append(float("nan"))
            continue
        fp.sum()
        f_pcs = np.cumsum(fp)
        t_pcs = np.cumsum(tp)
        prec = t_pcs / (f_pcs + t_pcs).astype(float)
        avg_prec = 0
        for i in range(submission_array.shape[0]):
            if tp[i]:
                avg_prec += prec[i]
        m_aps.append(avg_prec / n_pos.astype(float))
    m_aps = np.array(m_aps)
    m_ap = np.nanmean(m_aps)
    w_ap = m_aps * np.nansum(gt_array, axis=0) / gt_array.sum().sum().astype(float)
    return m_ap, w_ap, m_aps


def charades_map(submission_array, gt_array):
    """
    Approximate version of the charades evaluation function
    For precise numbers, use the submission file with the official matlab script
    """
    fix = submission_array.copy()
    empty = np.sum(gt_array, axis=1) == 0
    fix[empty, :] = np.NINF
    return mAP(fix, gt_array)


def charades_nanmap(submission_array, gt_array):
    """
    Approximate version of the charades evaluation function
    For precise numbers, use the submission file with the official matlab script
    """
    fix = submission_array.copy()
    empty = np.sum(gt_array, axis=1) == 0
    fix[empty, :] = np.NINF
    return nanmap(fix, gt_array)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

"""
Command line tool to download Ego4D datasets.

Examples:
      python ego4d/internal/validation/cli.py \
        -i "s3://ego4d-unict/metadata_v7" \
        -mf ego4d/internal/validation/standard_metadata/ego4d

      python ego4d/internal/validation/cli.py \
        -i "s3://ego4d-georgiatech/metadata_v5" \
        -mf ego4d/internal/validation/standard_metadata/ego4d \
        -u georgiatech \
        -o errors
"""
import boto3
import botocore.client as bclient
from ego4d.cli.universities import UNIV_TO_BUCKET
from ego4d.internal.s3 import get_client
from ego4d.internal.validation.config import Config, config_from_args, validate_config
from ego4d.internal.validation.validate import run_validation


def main_cfg(cfg: Config) -> None:

    validated_cfg = validate_config(cfg)
    run_validation(
        manifest_dir=validated_cfg.input_directory,
        input_university=validated_cfg.input_university,
        released_video_path=validated_cfg.released_video_path,
        standard_metadata_folder=validated_cfg.metadata_folder,
        output_dir=validated_cfg.output_dir,
        num_workers=validated_cfg.num_workers,
        expiry_time_sec=validated_cfg.expiry_time_sec,
        version=validated_cfg.version,
        skip_mp4_check=validated_cfg.skip_mp4_check,
    )


def main() -> None:
    config = config_from_args()
    main_cfg(config)


if __name__ == "__main__":
    main()



# Copyright (c) Meta Platforms, Inc. and affiliates. All Rights Reserved.

from dataclasses import dataclass
from typing import Tuple

from ego4d.features.config import BaseModelConfig, InferenceConfig
from ego4d.features.models.common import FeedVideoInput
from pytorchvideo.models.hub.vision_transformers import mvit_base_16, mvit_base_32x3
from pytorchvideo.transforms import ApplyTransformToKey, ShortSideScale
from torch.nn import Identity, Module
from torchvision.transforms import Compose, Lambda
from torchvision.transforms._transforms_video import CenterCropVideo, NormalizeVideo


@dataclass
class ModelConfig(BaseModelConfig):
    pretrained_dataset: str = "k400"

    # transformation config
    side_size: int = 256
    crop_size: int = 224
    mean: Tuple[float] = (0.45, 0.45, 0.45)
    std: Tuple[float] = (0.225, 0.225, 0.225)


def load_model(
    inference_config: InferenceConfig,
    config: ModelConfig,
    patch_final_layer: bool = True,
) -> Module:
    assert config.pretrained_dataset in ("k400", "imagenet")
    if config.pretrained_dataset == "k400":
        print("Loading K400 MViT")
        model = mvit_base_32x3(pretrained=True)
    else:
        print("Loading ImageNet MViT")
        model = mvit_base_16(pretrained=True)

    assert model is not None

    if patch_final_layer:
        model.head = Identity()

    # Set to GPU or CPU
    model = FeedVideoInput(model)
    model = model.eval()
    model = model.to(inference_config.device)
    return model


def get_transform(inference_config: InferenceConfig, config: ModelConfig):
    transforms = [
        Lambda(lambda x: x / 255.0),
        NormalizeVideo(config.mean, config.std),
        ShortSideScale(size=config.side_size),
        CenterCropVideo(config.crop_size),
    ]
    # image-based dataset
    if config.pretrained_dataset == "imagenet":
        # NOTE untested due to MViT imagenet not not available on torch hub
        transforms += [Lambda(lambda x: x.squeeze_(2))]
    return Compose(
        [
            ApplyTransformToKey(
                key="video",
                transform=Compose(transforms),
            )
        ]
    )



import json
import os
import shutil
import subprocess
import tempfile
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

import av

import boto3
import cv2
import hydra
import numpy as np
import pandas as pd

from ego4d.internal.s3 import StreamPathMgr

from iopath.common.file_io import PathManager
from iopath.common.s3 import S3PathHandler

from omegaconf import OmegaConf

from tqdm.auto import tqdm

stream_pathmgr = StreamPathMgr()

pathmgr = PathManager()
pathmgr.register_handler(S3PathHandler(profile="default"))


VRS_BIN = "/private/home/miguelmartin/repos/vrs/build/tools/vrs/vrs"
COLMAP_BIN = "/private/home/miguelmartin/repos/colmap/build/src/exe/colmap"


@dataclass
class ColmapConfig:
    in_metadata_path: Optional[str]
    in_videos: Optional[Dict[str, str]]  # device_id -> path
    output_dir: str
    download_video_files: bool

    sync_exo_views: bool
    rot_mode: int

    camera_model: str
    include_aria: bool

    mobile_frames: Optional[List[int]] = None
    aria_frames: Optional[Dict[str, List[int]]] = None
    aria_walkthrough_start_sec: Optional[float] = None
    aria_walkthrough_end_sec: Optional[float] = None
    aria_use_sync_info: bool = False

    exo_frames: Optional[Dict[str, List[int]]] = None
    exo_from_frame: Optional[int] = None
    exo_to_frame: Optional[int] = None
    frame_rate: Optional[float] = 1 / 2
    name: Optional[str] = None
    aria_fps: float = 30
    exo_fps: Optional[float] = None
    mobile_fps: Optional[float] = None

    run_colmap: bool = False
    colmap_bin: Optional[str] = None
    vrs_bin: Optional[str] = None

    force_download: Optional[bool] = False
    take_id: Optional[str] = None
    video_source: Optional[str] = None


def get_uniq_cache_root_dir(config) -> str:
    assert (
        config.video_source is not None and config.take_id is not None
    ), "need video source and take_id to set unique output dir for COLMAP data"
    assert config.output_dir is not None, "need cache root dir"
    return os.path.join(
        config.output_dir, "cache", f"{config.video_source}_{config.take_id}"
    )


def get_timesync_path(config) -> str:
    return os.path.join(get_uniq_cache_root_dir(config), "timesync.csv")


def pilot_video_metadata(uni_name: str, take_id: str) -> str:
    return f"s3://ego4d-consortium-sharing/internal/egoexo_pilot/{uni_name}/{take_id}/metadata.json"  # noqa


def get_colmap_data_dir(config) -> str:
    assert (
        config.video_source is not None and config.take_id is not None
    ), "need video source and take_id to set unique output dir for COLMAP data"
    name = config.name
    if name is None:
        name = f"{config.camera_model}_s{int(config.sync_exo_views)}_r{config.rot_mode}_a{int(config.include_aria)}_fr{config.frame_rate:.3f}"  # noqa
    return os.path.join(
        config.output_dir, "output", f"{config.video_source}_{config.take_id}", name
    )


def frames_for_region(start_frame: int, end_frame: int, skip_frames: int) -> List[int]:
    # [start, end]
    return list(range(start_frame, end_frame + 1))[0::skip_frames]


def get_fps(fp: str) -> float:
    with av.open(stream_pathmgr.open(fp)) as cont:
        return float(cont.streams[0].average_rate)


def produce_colmap_script(
    config: ColmapConfig,
    skip_frame_gen: bool = False,
) -> ColmapConfig:
    vrs_bin = config.vrs_bin
    colmap_bin = config.colmap_bin
    if vrs_bin is None:
        vrs_bin = VRS_BIN
    if colmap_bin is None:
        colmap_bin = COLMAP_BIN

    if not os.path.exists(colmap_bin) and config.run_colmap:
        raise AssertionError(f"colmap binary does not exist: {colmap_bin}")

    if not os.path.exists(vrs_bin):
        raise AssertionError(f"vrs binary does not exist: {colmap_bin}")

    print("Getting data ready...")
    _get_data_ready(config, config.force_download, vrs_bin, skip_frame_gen)
    print(f"Writing colmap script using {colmap_bin}")
    _write_colmap_file(config, colmap_bin)
    print(f"Please run {os.path.join(get_colmap_data_dir(config), 'run_colmap.sh')}")
    return config


def _get_data_ready(
    config: ColmapConfig, force_download: bool, vrs_bin: str, skip_frame_gen: bool
):
    by_dev_id = _download_data(config, force_download)
    if not skip_frame_gen:
        _extract_frames(config, by_dev_id, vrs_bin)


def download_andor_generate_streams(
    metadata: Dict[str, Any],
    download_video_files: bool,
    force_download: bool,
    output_dir: str,
) -> Dict[str, Any]:
    os.makedirs(output_dir, exist_ok=True)
    if metadata["timesync_csv_path"] is not None:
        local_timesync_csv = os.path.join(output_dir, "timesync.csv")
        if metadata["timesync_csv_path"].startswith("s3://"):
            _download_s3_path(metadata["timesync_csv_path"], local_timesync_csv)
        else:
            # local path
            shutil.copy2(metadata["timesync_csv_path"], local_timesync_csv)

    by_dev_id = {}
    for v in tqdm(metadata["videos"]):
        by_dev_id[v["device_id"]] = v
        s3_path = v["s3_path"]
        if s3_path.startswith("s3://") and (
            download_video_files or "aria" in v["device_id"]
        ):
            local_path = os.path.join(output_dir, v["device_id"])
            by_dev_id[v["device_id"]]["local_path"] = local_path
            if os.path.exists(local_path) and not force_download:
                continue
            print(f"Fetching {s3_path} to {local_path}")
            _download_s3_path(s3_path, local_path)
        else:
            by_dev_id[v["device_id"]]["local_path"] = s3_path
    return by_dev_id


def _download_data(config: ColmapConfig, force_download: bool) -> Dict[str, Any]:
    assert config.output_dir is not None, "Please set the output directory"

    if config.in_metadata_path is None:
        assert (
            config.in_videos is not None
        ), "require input video paths if metdata path is None"
        assert not config.sync_exo_views, "cannot sync videos without metadata"

        print("WARNING: using input video paths. This should only be used for TESTING")
        meta = {"videos": [], "timesync_csv_path": None}
        for dev_id, path in config.in_videos.items():
            meta["videos"].append(
                {
                    "device_id": dev_id,
                    "s3_path": path,  # may actually be a local path
                }
            )
    else:
        assert (
            config.in_videos is None
        ), "cannot use input videos if metadata is not None"
        meta = json.load(pathmgr.open(config.in_metadata_path))
        if config.take_id is None or config.video_source is None:
            config.take_id = meta["take_id"]
            config.video_source = meta["video_source"]

    output_dir = get_uniq_cache_root_dir(config)
    return download_andor_generate_streams(
        metadata=meta,
        output_dir=output_dir,
        download_video_files=config.download_video_files,
        force_download=force_download,
    )


def _extract_frames(config, by_dev_id, vrs_bin):
    frame_out_dir = os.path.join(get_colmap_data_dir(config), "frames")
    shutil.rmtree(frame_out_dir, ignore_errors=True)
    os.makedirs(frame_out_dir, exist_ok=True)

    aria_keys = [k for k in by_dev_id.keys() if "aria" in k]
    exo_keys = [k for k in by_dev_id.keys() if "cam" in k]

    synced_df = None
    if config.sync_exo_views or (
        config.aria_walkthrough_start_sec is None
        and config.aria_walkthrough_end_sec is None
        and config.in_metadata_path is not None
    ):
        assert (
            config.aria_walkthrough_start_sec is None
        ), "If not providing end sec of walkthrough time, please set start to none for timesync option"  # noqa
        timesync_df = pd.read_csv(get_timesync_path(config))
        cam01_idx = timesync_df.cam01_global_time.first_valid_index()
        cam02_idx = timesync_df.cam02_global_time.first_valid_index()
        cam03_idx = timesync_df.cam03_global_time.first_valid_index()
        cam04_idx = timesync_df.cam04_global_time.first_valid_index()
        fsf_idx = max(cam01_idx, cam02_idx, cam03_idx, cam04_idx)  # first synced frame
        synced_df = timesync_df.iloc[fsf_idx:]

    if config.include_aria:
        if config.aria_frames is None:
            config.aria_frames = {k: None for k in aria_keys}
        print("aria used", aria_keys)
        for k in aria_keys:
            print(k, config.aria_walkthrough_start_sec, config.aria_walkthrough_end_sec)
            with tempfile.TemporaryDirectory() as tempdir:
                aria_frame_path = os.path.join(tempdir, "aria01_frames")
                os.makedirs(aria_frame_path, exist_ok=True)
                extract_aria_frames(
                    vrs_bin=vrs_bin,
                    aria_vrs_file_path=by_dev_id[k]["local_path"],
                    out_dir=aria_frame_path,
                    synced_df=synced_df,
                    from_point=config.aria_walkthrough_start_sec,
                    to_point=config.aria_walkthrough_end_sec,
                )
                skip_aria = (
                    int(np.round(config.aria_fps * config.frame_rate))
                    if config.aria_fps is not None and config.frame_rate is not None
                    else None
                )
                subsample_aria_frames(
                    aria_frame_path,
                    skip_aria,
                    frame_out_dir,
                    config.aria_frames[k],
                    config,
                    aria_key=k,
                )

    exo_fps = get_fps(by_dev_id["cam01"]["local_path"])
    pts_by_key = {}
    if config.sync_exo_views:
        assert set(exo_keys) == {"cam01", "cam02", "cam03", "cam04"}, (
            "for sync_mode, exo cameras needed: {cam01, cam02, cam03, cam04}, but got "
            + f"{set(exo_keys)}"
        )
        f1 = config.exo_from_frame
        f2 = config.exo_to_frame
        skip_num_frames = int(np.round(exo_fps * config.frame_rate))
        if config.exo_frames is not None:
            raise AssertionError("cannot use exo_frames in sync mode")
        cam01_pts = [
            int(x) for x in synced_df.cam01_pts.iloc[f1:f2:skip_num_frames].tolist()
        ]
        cam02_pts = [
            int(x) for x in synced_df.cam02_pts.iloc[f1:f2:skip_num_frames].tolist()
        ]
        cam03_pts = [
            int(x) for x in synced_df.cam03_pts.iloc[f1:f2:skip_num_frames].tolist()
        ]
        cam04_pts = [
            int(x) for x in synced_df.cam04_pts.iloc[f1:f2:skip_num_frames].tolist()
        ]
        pts_by_key["cam01"] = cam01_pts
        pts_by_key["cam02"] = cam02_pts
        pts_by_key["cam03"] = cam03_pts
        pts_by_key["cam04"] = cam04_pts
    else:
        if config.exo_from_frame is not None:
            assert (
                config.exo_frames is None
            ), "if given exo_from_frame and exo_to_frame - you cannot supply frames"
            assert config.exo_to_frame is not None, "need to_frame"
            assert config.frame_rate is not None, "need frame rate"
            f1 = config.exo_from_frame
            f2 = config.exo_to_frame
            skip_num_frames = int(np.round(exo_fps * config.frame_rate))

            frames = list(range(f1, f2))[::skip_num_frames]
            config.exo_frames = {k: frames for k in exo_keys}

        for k in exo_keys:
            assert config.exo_frames[k] is not None, f"exo_frames have frames for {k}"
            print(f"{k}_frames=\n{config.exo_frames[k]}")
            pts = get_all_pts(by_dev_id[k]["local_path"], idx_set=config.exo_frames[k])
            pts_by_key[k] = pts

    mobile_frames = config.mobile_frames
    if mobile_frames is not None:
        mobile_pts = get_all_pts(by_dev_id["mobile"]["local_path"], mobile_frames)
    else:
        mobile_pts = get_all_pts(by_dev_id["mobile"]["local_path"])
        mobile_fps = (
            get_fps(by_dev_id["mobile"]["local_path"])
            if config.mobile_fps is None
            else config.mobile_fps
        )
        msf = int(np.round(mobile_fps * config.frame_rate))

        nf = len(mobile_pts)
        config.mobile_frames = list(range(0, nf))[::msf]
        mobile_pts = get_all_pts(
            by_dev_id["mobile"]["local_path"], config.mobile_frames
        )

    pts_by_key["mobile"] = mobile_pts
    for dev_name, pts_set in tqdm(pts_by_key.items()):
        assert issorted(pts_set)
        print(dev_name, len(pts_set))
        frames = get_frames_list(by_dev_id[dev_name]["local_path"], pts_set)

        save_off_frames(
            frames, pts_set, f"{frame_out_dir}/{dev_name}", config.rot_mode == 2
        )


def _write_colmap_file(config: ColmapConfig, colmap_bin: str):
    colmap_bin_str = "${1:-" + colmap_bin + "}"
    gpu_idx_str = "${2:-0}"
    script = f"""
COLMAP_BIN={colmap_bin_str}
GPU_IDX={gpu_idx_str}
echo "Using $COLMAP_BIN"
echo "You can provide your own COLMAP binary by providing it as the first arg to this script"

echo "Using GPU: $GPU_IDX"
echo "You can provide -1 for CPU or another index as the second arg to this script"

SCRIPT_DIR=$(dirname -- $0)
DB_PATH=$SCRIPT_DIR/colmap.db
IN_FRAME_DIR=$SCRIPT_DIR/frames
MODEL_DIR=$SCRIPT_DIR/colmap_model
UNDIST_DIR=$SCRIPT_DIR/undistorted_images

mkdir $MODEL_DIR
mkdir $UNDIST_DIR


$COLMAP_BIN feature_extractor \\
   --database_path $DB_PATH \\
   --image_path $IN_FRAME_DIR \\
   --SiftExtraction.gpu_index $GPU_IDX \\
   --ImageReader.single_camera_per_folder 1 \\
   --ImageReader.camera_mode {config.camera_model}

$COLMAP_BIN exhaustive_matcher --database_path $DB_PATH

$COLMAP_BIN mapper \\
    --database_path $DB_PATH \\
    --image_path $IN_FRAME_DIR \\
    --output_path $MODEL_DIR \\
    --Mapper.ba_global_pba_gpu_index $GPU_IDX

$COLMAP_BIN model_converter \
    --input_path $MODEL_DIR/0 \
    --output_path $MODEL_DIR/0 \
    --output_type TXT

$COLMAP_BIN image_undistorter \\
    --image_path $IN_FRAME_DIR \\
    --input_path $MODEL_DIR/0 \\
    --output_path $SCRIPT_DIR \\
    --output_type COLMAP \\
    --max_image_size 2000

$COLMAP_BIN model_analyzer --path $MODEL_DIR/0 > $SCRIPT_DIR/analysis_0.txt
"""
    with open(os.path.join(get_colmap_data_dir(config), "run_colmap.sh"), "w") as out_f:
        out_f.write(script)

    with open(os.path.join(get_colmap_data_dir(config), "config.yaml"), "w") as out_f:
        out_f.write(OmegaConf.to_yaml(config))
    print("Config=")
    print(OmegaConf.to_yaml(config))


def issorted(xs):
    return sorted(xs) == xs


def get_frames_iter(video_path, pts_set):
    min_pts = min(pts_set)
    max_pts = max(pts_set)

    with av.open(stream_pathmgr.open(video_path)) as cont:
        v_stream = cont.streams.video[0]
        cont.seek(min_pts, stream=v_stream)
        for frame in cont.decode(video=0):
            if frame.pts > max_pts:
                break
            if frame.pts not in pts_set:
                continue
            if isinstance(frame, av.VideoFrame):
                x = frame.to_ndarray(format="rgb24")
                yield frame.pts, x


def get_frames_list(video_path, pts_set):
    min_pts = min(pts_set)
    max_pts = max(pts_set)

    frame_data = {}
    with av.open(stream_pathmgr.open(video_path)) as cont:
        v_stream = cont.streams.video[0]
        cont.seek(min_pts, stream=v_stream)
        for frame in cont.decode(video=0):
            if frame.pts > max_pts:
                break
            if frame.pts not in pts_set:
                continue
            if isinstance(frame, av.VideoFrame):
                x = frame.to_ndarray(format="rgb24")
                frame_data[frame.pts] = x

    assert len(frame_data) == len(pts_set)
    return frame_data


def save_off_frames(frames, pts_set, out_dir, rot90):
    os.makedirs(out_dir, exist_ok=True)
    for idx, pts in tqdm(enumerate(pts_set), total=len(pts_set)):
        frame_data = frames[pts]
        out_path = os.path.join(out_dir, f"{idx}.jpg")

        img = cv2.cvtColor(frame_data, cv2.COLOR_RGB2BGR)
        if rot90:
            img = cv2.rotate(img, cv2.ROTATE_90_COUNTERCLOCKWISE)
        cv2.imwrite(out_path, img)


def get_all_pts(video_path, idx_set=None):
    with av.open(stream_pathmgr.open(video_path)) as cont:
        frame_pts = []
        idx = 0
        for frame in cont.demux(video=0):
            if frame.pts is None:
                continue
            if idx_set is not None and idx not in idx_set:
                idx += 1
                continue
            frame_pts.append(frame.pts)
            idx += 1
        return frame_pts


def extract_aria_frames(
    vrs_bin: str,
    aria_vrs_file_path: str,
    out_dir: str,
    synced_df: Optional[pd.DataFrame],
    stream_name="214-1",
    from_point: Optional[float] = None,
    to_point: Optional[float] = None,
):
    if synced_df is not None:
        assert from_point is None and to_point is None
        # TODO: get a better estimate for this
        t_end = synced_df["aria01_1201-2_capture_timestamp_ns"].iloc[0] / 10**9
        # !$vrs extract-images $data_path/aria01.mp4 + "214-1
        # " --before $t_end --to $input_aria_frames
        cmd = [
            vrs_bin,
            "extract-images",
            aria_vrs_file_path,
            "+",
            stream_name,
            "--before",
            f"{t_end}",
            "--to",
            out_dir,
        ]
        print("Running:")
        print(" ".join(cmd))
        subprocess.run(cmd)
    else:
        assert (
            to_point is not None or from_point is not None
        ), "will not exact all aria frames, please specify a region"
        cmd = [
            vrs_bin,
            "extract-images",
            aria_vrs_file_path,
            "+",
            stream_name,
        ]
        if to_point is not None:
            cmd += [
                "--before",
                f"{to_point}",
            ]
        if from_point is not None:
            cmd += ["--after", f"{from_point}"]

        cmd += ["--to", out_dir]
        print("Running:")
        print(" ".join(cmd))
        subprocess.run(cmd)


def subsample_aria_frames(
    input_aria_frame_dir,
    skip_num_frames,
    frame_out_dir,
    aria_frames_subset,
    config,
    aria_key,
):
    aria_frames = os.listdir(input_aria_frame_dir)
    out_frame_dir = os.path.join(frame_out_dir, "aria01")
    shutil.rmtree(out_frame_dir, ignore_errors=True)
    os.makedirs(out_frame_dir, exist_ok=True)

    f_idx = aria_frames_subset
    if f_idx is None:
        f_idx = list(range(len(aria_frames)))[::skip_num_frames]

    config.aria_frames[aria_key] = f_idx

    for idx in f_idx:
        f = aria_frames[idx]
        full_path = os.path.join(input_aria_frame_dir, f)
        out_path = os.path.join(out_frame_dir, f)
        if config.rot_mode == 1:
            img = cv2.imread(full_path)
            img = cv2.rotate(img, cv2.ROTATE_90_CLOCKWISE)
            cv2.imwrite(out_path, img)
        else:
            shutil.copy(full_path, out_path)


def _download_s3_path(s3_path: str, out_path: str):
    assert s3_path.startswith("s3://")
    s3 = boto3.client("s3")
    no_s3_path = s3_path.split("s3://")[1]
    temp = no_s3_path.split("/")
    bucket_name = temp[0]
    key = "/".join(temp[1:])
    s3.download_file(bucket_name, key, out_path)


def create_video(video_path, pts_set, out_path, fps=30, codec="h264", flip_90=True):
    """
    This is used to construct a video, useful for feeding into EasyMocap
    """
    frames = get_frames_iter(video_path, pts_set)
    print("Received frames")

    with av.open(stream_pathmgr.open(video_path)) as in_container:
        in_video_stream = in_container.streams.video[0]
        width, height = in_video_stream.width, in_video_stream.height
        if width % 2 != 0:
            width -= 1
        elif height % 2 != 0:
            height -= 1
        pix_fmt = in_video_stream.pix_fmt

    v_frame = 0
    with av.open(out_path, mode="w") as out_cont:
        video_stream = out_cont.add_stream(codec, fps)
        if flip_90:
            video_stream.width = height
            video_stream.height = width
        else:
            video_stream.width = width
            video_stream.height = height
        video_stream.pix_fmt = pix_fmt
        out_cont.start_encoding()
        for _, frame in frames:
            img = frame[0:height, 0:width, :]
            if flip_90:
                img = cv2.rotate(img, cv2.ROTATE_90_COUNTERCLOCKWISE)

            copy_frame = av.VideoFrame.from_ndarray(img, format="rgb24")
            for packet in video_stream.encode(copy_frame):
                out_cont.mux(packet)
                v_frame += 1
    print(f"encoded {v_frame} frames, from given {len(pts_set)} frames")
    return v_frame


@hydra.main(config_path="configs", config_name=None)
def run_colmap_preprocess(config: ColmapConfig):
    print(f"Working Directory {os.getcwd()} - please make sure you're using abs paths")
    ret = produce_colmap_script(config)
    if config.run_colmap:
        subprocess.run(
            ["bash", os.path.join(get_colmap_data_dir(config), "run_colmap.sh")]
        )
    return ret


if __name__ == "__main__":
    run_colmap_preprocess()  # pyre-ignore



import numpy as np
import torch
import torch.nn as nn

from ego4d.research.clep.config import ModelConfig


def _get_layers(initial_dim, config):
    return [
        nn.Linear(initial_dim, config.final_proj_size),
        nn.ReLU(True),
        nn.Linear(config.final_proj_size, config.final_proj_size),
    ]


class EgoLangaugeAssociation(nn.Module):
    def __init__(self, config: ModelConfig):
        super().__init__()

        self.config = config
        txt_layers = _get_layers(config.nlp_feature_size, config)
        viz_layers = _get_layers(config.visual_feature_size, config)
        self.text_proj = nn.Sequential(*tuple(txt_layers))
        self.visual_proj = nn.Sequential(*tuple(viz_layers))

        self.apply(self.init_weights)

        # don't want to init this with 0
        self.logit_scale = nn.Parameter(torch.ones([]) * np.log(1 / 0.07))

    def forward(self, x):
        ve = self.visual_proj(x["video"])
        te = self.text_proj(x["text"])
        return ve, te, self.logit_scale

    def init_weights(self, module):
        if isinstance(module, nn.Linear):
            torch.nn.init.xavier_uniform_(
                module.weight.data, gain=torch.nn.init.calculate_gain("relu")
            )
            module.bias.data.zero_()



# Copyright (c) Meta Platforms, Inc. and affiliates. All Rights Reserved.

from dataclasses import dataclass
from typing import Any, Dict

import torch
from ego4d.features.config import BaseModelConfig, InferenceConfig
from speechbrain.pretrained import EncoderDecoderASR
from torch.nn import Module
from torchvision.transforms import Compose


@dataclass
class ModelConfig(BaseModelConfig):
    source: str = "speechbrain/asr-crdnn-transformerlm-librispeech"
    savedir: str = "pretrained_models/asr-crdnn-transformerlm-librispeech"


class WrapAsrModel(Module):
    def __init__(self, model):
        super().__init__()
        self.model = model

    def forward(self, x) -> Dict[str, Any]:
        assert len(x["audio_sample_rate"]) == 1
        if "audio" not in x:
            return []

        # see
        # https://github.com/speechbrain/speechbrain/blob/develop/speechbrain/pretrained/interfaces.py#L545
        # docstring is out-dated, need to use `audio_normalizer` not `normalizer`
        inp = self.model.audio_normalizer(
            x["audio"].permute(1, 0), x["audio_sample_rate"].item()
        )

        wavs = inp.unsqueeze(0)
        wav_lens = torch.tensor([1.0])
        # https://github.com/speechbrain/speechbrain/blob/598f6eda70f9b0c9ad49b393114ff483add1fd25/speechbrain/pretrained/interfaces.py#L595
        with torch.no_grad():
            wav_lens = wav_lens.to(self.model.device)
            encoder_out = self.model.encode_batch(wavs, wav_lens)
            predicted_tokens, scores = self.model.mods.decoder(encoder_out, wav_lens)
            pred_words = [
                self.model.tokenizer.decode_ids(token_seq)
                for token_seq in predicted_tokens
            ]
        # or...
        # pred_words, pred_tokens = self.model(wavs, wav_lens)

        return {
            "text": pred_words[0] if len(pred_words[0]) > 0 else None,
            "score": scores[0].item(),
        }


def load_model(
    inference_config: InferenceConfig,
    config: ModelConfig,
    patch_final_layer: bool = True,
) -> Module:
    if patch_final_layer:
        print("WARNING: this model outputs text, and patching is not supporting")
    model = EncoderDecoderASR.from_hparams(
        source=config.source,
        savedir=config.savedir,
        run_opts={"device": inference_config.device},
    )
    return WrapAsrModel(model)


def get_transform(inference_config: InferenceConfig, config: ModelConfig):
    # do nothing
    return Compose([])



# pyre-unsafe
import logging
import os
import re
from dataclasses import dataclass
from typing import List, Optional, Tuple

import botocore


@dataclass
class FileInfo:
    key: str
    size: int


def is_file_readable(file_path):
    return os.path.isfile(file_path) and os.access(file_path, os.R_OK)


def parse_bucket_key(key: str) -> Tuple[str, str]:
    # Determine from the key
    assert key and key.startswith(
        "s3://"
    ), f"Invalid key without bucket supplied: {key}"
    m = re.match("^s3://([^/]*)/(.*)$", key)
    assert m, f"Invalid s3:// search key: {key}"
    grp = m.groups()
    assert len(grp) == 2
    return grp[0], grp[1]


class S3Helper:
    _bucket_name = None

    def __init__(self, s3, bucket_name):
        self._bucket_name = bucket_name
        # self._s3
        self._s3 = s3

    @property
    def bucket(self):
        return self._bucket_name

    def ls(self, prefix: str, max_results=-1, **kwargs) -> Tuple[bool, List[FileInfo]]:
        if max_results < 0:
            paginator = self._s3.get_paginator("list_objects_v2")
            pages = paginator.paginate(Bucket=self._bucket_name, Prefix=prefix)
            results = []
            for page in pages:
                for f in page["Contents"]:
                    results.append(FileInfo(f["Key"], f["Size"]))
            return False, results
        else:
            # legacy
            ls_result = self._s3.list_objects_v2(
                Bucket=self._bucket_name, Prefix=prefix, MaxKeys=max_results, **kwargs
            )
            return (
                ls_result["IsTruncated"],
                [FileInfo(f["Key"], f["Size"]) for f in ls_result["Contents"]],
            )

    def get_file(self, key: str, local_path: str, **kwargs) -> Optional[str]:
        print(f"Downloading: {key} to {local_path}")
        try:
            self._s3.download_file(
                Bucket=self._bucket_name, Key=key, Filename=local_path, **kwargs
            )
        except botocore.exceptions.ClientError as e:
            # If a client error is thrown, then check that it was a 404 error.
            # If it was a 404 error, then the file does not exist.
            error_code = int(e.response["Error"]["Code"])
            if error_code == 404:
                return None
            else:
                raise e
        return local_path

    def exists(self, key: str, bucket: Optional[str] = None, **kwargs) -> bool:
        if not bucket:
            bucket, key = parse_bucket_key(key)
            assert key and bucket
        try:
            obj = self._s3.get_object(Bucket=bucket, Key=key)
            return bool(obj)
        except botocore.exceptions.ClientError as e:
            logging.exception("AWS get_size exception: ", e)
            raise e



# pyre-unsafe

import json
import random
import subprocess
import time
from dataclasses import dataclass
from fractions import Fraction
from typing import Any, Dict, List, Optional, Tuple

from ego4d.internal.validation.manifest import Error, ErrorLevel


@dataclass(frozen=True)
class VideoInfo:
    fps: Fraction
    sar: Fraction
    dar: Optional[Fraction]
    sample_width: int
    sample_height: int
    vcodec: Optional[str] = None
    acodec: Optional[str] = None
    vstart: Optional[float] = None
    astart: Optional[float] = None
    vduration: Optional[float] = None
    aduration: Optional[float] = None
    mp4_duration: Optional[float] = None
    video_time_base: Optional[Fraction] = None
    audio_sample_rate: Optional[int] = None
    audio_channel_layout: Optional[str] = None
    rotate: Optional[int] = None
    unstructured_stream_data: Optional[Dict[str, Any]] = None

    @property
    def display_height(self) -> int:
        return self.sample_height

    @property
    def display_width(self) -> int:
        assert self.sar is not None
        return int(self.sample_width * (self.sar.numerator / self.sar.denominator))


def get_video_info(
    filename: str,
    name: Optional[str] = None,
) -> Tuple[Optional[VideoInfo], Optional[Error]]:
    """
    Return:
        VideoInfo: information of the video stored in bucket_name specified by the object_name
        If error, returns None and log the error.
    """
    name = name if name is not None else filename

    cmd = [
        "ffprobe",
        "-i",
        f"{filename}",
        "-v",
        "error",
        "-print_format",
        "json",
        "-show_format",
        "-show_streams",
        "-hide_banner",
    ]

    def to_fraction(fps):
        if fps is None:
            return 0
        a, b = fps.split("/")
        if int(b) == 0:
            return 0
        return Fraction(int(a), int(b))

    result = None
    x = 1
    for _ in range(10):
        time.sleep(random.random() * x + x / 2)
        try:
            result = subprocess.run(cmd, encoding="utf-8", capture_output=True)
        except Exception as ex:
            return None, Error(
                ErrorLevel.ERROR,
                name,
                "ffmpeg_cannot_read_error",
                f"{ex}",
            )

        if result.stderr:
            if "tls" in result.stderr:
                x *= 2.5
                continue
            return None, Error(
                ErrorLevel.ERROR,
                name,
                "ffmpeg_cannot_read_error",
                f"{result.stderr}",
            )

        break

    assert result is not None

    # return result
    data = json.loads(result.stdout)

    fps = None
    width = None
    height = None
    sar = Fraction(1, 1)
    dar = None
    vcodec = None
    acodec = None
    vstart = None
    astart = None
    vduration = None
    aduration = None
    audio_sample_rate = None
    channel_layout = None
    rotate = None
    vtb = None
    mp4_duration = None

    if "format" in data:
        if "duration" in data["format"]:
            mp4_duration = float(data["format"]["duration"])

    # it's possible for some files
    # to not have this information available
    if "streams" in data:
        for stream in data["streams"]:
            if "avg_frame_rate" in stream:
                s_fps = to_fraction(stream["avg_frame_rate"])
                if fps is None or fps < s_fps:
                    fps = s_fps

            if "sample_aspect_ratio" in stream:
                a, b = stream["sample_aspect_ratio"].split(":")
                sar = Fraction(int(a), int(b))

            if "display_aspect_ratio" in stream:
                a, b = stream["display_aspect_ratio"].split(":")
                dar = Fraction(int(a), int(b))

            if "codec_name" in stream:
                if "width" in stream:
                    vcodec = stream["codec_name"]
                    width = stream["width"]
                else:
                    acodec = stream["codec_name"]

            if "height" in stream:
                height = stream["height"]

            if "duration" in stream:
                if "width" in stream:
                    vduration = float(stream["duration"])
                else:
                    aduration = float(stream["duration"])

            if "start_time" in stream:
                if "width" in stream:
                    vstart = float(stream["start_time"])
                else:
                    astart = float(stream["start_time"])

            if "time_base" in stream and "width" in stream:
                vtb = to_fraction(stream["time_base"])

            if "sample_rate" in stream:
                audio_sample_rate = int(stream["sample_rate"])

            if "channel_layout" in stream:
                channel_layout = stream["channel_layout"]

            if "tags" in stream:
                if "rotate" in stream["tags"]:
                    rotate = int(stream["tags"]["rotate"])

    video_info = VideoInfo(
        sar=sar,
        dar=dar,
        sample_width=width,
        sample_height=height,
        vcodec=vcodec,
        acodec=acodec,
        vduration=vduration,
        aduration=aduration,
        vstart=vstart,
        astart=astart,
        fps=fps,
        audio_sample_rate=audio_sample_rate,
        audio_channel_layout=channel_layout,
        rotate=rotate,
        unstructured_stream_data=data,
        video_time_base=vtb,
        mp4_duration=mp4_duration,
    )
    return video_info, None



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

import json
import tempfile

import pytest

from .config import config_from_args


def test_invalid_flags():
    with pytest.raises(RuntimeError):
        # missing required flag output_directory
        config_from_args(["--version=v1", "--datasets", "full_scale"])

    # with pytest.raises(RuntimeError):
    #     # missing required flag datasets
    #     config_from_args(["--version=v1", "--output_directory=~/test"])

    with pytest.raises(SystemExit):
        # invalid university choice
        config_from_args(
            [
                "--datasets",
                "full_scale",
                "--output_directory=~/test",
                "--version=v1",
                "--universities",
                "cmu",
                "fake",
            ]
        )

    with pytest.raises(SystemExit):
        # unrecognized flag
        config_from_args(
            [
                "--datasets",
                "full_scale",
                "--output_directory=~/test",
                "--version=v1",
                "--fake_flag=123",
            ]
        )

    with pytest.raises(SystemExit):
        # unrecognized flag
        config_from_args(
            [
                "--datasets",
                "full_scale",
                "--output_directory=~/test",
                "--version=v1",
                "--video_uids",
                "123",
                "456",
                "--video_uid_file=~/uid_file",
            ]
        )


def test_valid_flags():
    c = config_from_args(
        [
            "--datasets",
            "full_scale",
            "--output_directory=~/test",
            "--version=v1",
            "--universities",
            "cmu",
            "indiana",
            "--aws_profile_name=ego4d",
            "--video_uids",
            "123",
            "456",
            "789",
            "--metadata",
        ]
    )

    assert c.version == "v1"
    assert c.datasets == ["full_scale"]
    assert c.universities == ["cmu", "indiana"]
    assert c.output_directory == "~/test"
    assert c.aws_profile_name == "ego4d"
    assert c.video_uids == ["123", "456", "789"]
    assert c.metadata == True


def test_version_flags():
    c = config_from_args(
        [
            "--datasets",
            "full_scale",
            "--output_directory=~/test",
            "--version=v2",
            "--universities",
            "cmu",
            "indiana",
            "--aws_profile_name=ego4d",
            "--video_uids",
            "123",
            "456",
            "789",
            "--metadata",
        ]
    )

    assert c.version == "v2"
    assert c.datasets == ["full_scale"]
    assert c.universities == ["cmu", "indiana"]
    assert c.output_directory == "~/test"
    assert c.aws_profile_name == "ego4d"
    assert c.video_uids == ["123", "456", "789"]
    assert c.metadata == True


def test_video_uid_file():
    uids = "123 456\n789"

    with tempfile.NamedTemporaryFile(mode="w") as f:
        f.write(uids)
        f.flush()

        c = config_from_args(
            [
                "--datasets",
                "full_scale",
                "--output_directory=~/test",
                "--version=v1",
                f"--video_uid_file={f.name}",
            ]
        )

        assert c.video_uids == ["123", "456", "789"]


def test_json_file():
    conf = {
        "version": "v1",
        "datasets": ["full_scale"],
        "universities": ["cmu", "indiana"],
        "output_directory": "~/test",
        "aws_profile_name": "ego4d",
        "video_uids": ["123", "456", "789"],
    }

    with tempfile.NamedTemporaryFile(mode="w") as f:

        json.dump(conf, f)
        f.flush()

        # JSON values should be loaded into config
        c = config_from_args([f"--config_path={f.name}"])
        assert c.version == "v1"
        assert c.datasets == ["full_scale"]
        assert c.universities == ["cmu", "indiana"]
        assert c.output_directory == "~/test"
        assert c.aws_profile_name == "ego4d"
        assert c.video_uids == ["123", "456", "789"]

        # JSON values should be overridden by flags that are also specified on the
        # command line
        c = config_from_args(
            [
                f"--config_path={f.name}",
                "--output_directory=~/test2",
                "--video_uids",
                "000",
                "000",
            ]
        )
        assert c.version == "v1"
        assert c.datasets == ["full_scale"]
        assert c.universities == ["cmu", "indiana"]
        assert c.output_directory == "~/test2"
        assert c.aws_profile_name == "ego4d"
        assert c.video_uids == ["000", "000"]

        # If video_uid_file specified on command line, but video_uids is specified in
        # JSON file, then CLI should error
        with pytest.raises(RuntimeError):
            config_from_args(
                [
                    f"--config_path={f.name}",
                    "--output_directory=~/test2",
                    "--video_uid_file=~/uid_file",
                ]
            )



# Copyright (c) Meta Platforms, Inc. and affiliates. All Rights Reserved.

import functools
from fractions import Fraction
from typing import Any, Dict, List, Optional

import av
import av.error
import numpy as np
import torch
from ego4d.features.config import FeatureExtractConfig, get_transform, Video
from ego4d.research.dataset import VideoDataset

from ego4d.research.readers import PyAvReader, TorchAudioStreamReader
from pytorchvideo.data import UniformClipSampler
from pytorchvideo.data.encoded_video import EncodedVideo
from pytorchvideo.data.utils import thwc_to_cthw
from pytorchvideo.transforms import ApplyTransformToKey, ShortSideScale
from torch.utils.data import DataLoader
from torchvision.transforms import Compose
from tqdm.auto import tqdm


def get_frames(container, t1, t2, buffer, max_buffer_size, frame_window_size):
    # [t1, t2]
    ret = []

    tb = container.streams.video[0].time_base

    def is_in_range(frame):
        t = frame.pts * tb
        return t >= t1 and t < t2

    def exceeds_range(frame):
        return frame.pts * tb >= t2

    for frame in buffer.values():
        if is_in_range(frame):
            ret.append(frame)

    ret.sort(key=lambda x: x.pts)

    if len(ret) == 0 or ret[-1].pts + 1 < t2 / tb:
        while True:
            try:
                for frame in container.decode(video=0):
                    if frame.pts is None:
                        raise AssertionError("frame is None")
                    if not isinstance(frame, av.VideoFrame):
                        raise AssertionError("other packets not supported")

                    buffer[frame.pts] = frame

                    if len(buffer) > max_buffer_size:
                        del buffer[min(buffer.keys())]

                    if is_in_range(frame):
                        ret.append(frame)
                    elif exceeds_range(frame):
                        break
                break
            except av.error.EOFError:
                seek_pts = buffer[max(buffer.keys())].pts
                container.seek(seek_pts, stream=container.streams.video[0])

        ret.sort(key=lambda x: x.pts)

    ret = list({frame.pts: frame for frame in ret}.values())
    pts_in_ret = [frame.pts for frame in ret]
    if not (np.diff(pts_in_ret) > 0).all():
        raise AssertionError("not increasing sequence of frames")
    assert len(ret) == frame_window_size, f"{len(ret)} != {frame_window_size}"
    return ret


class EncodedVideoCached:
    def __init__(self, path, frame_buffer_size=16):
        self.path = path
        self.vid = EncodedVideo.from_path(path, decoder="pyav")
        self.vid._container.seek(0)

        self.frame_buffer_size = frame_buffer_size
        self.frame_buffer = {}
        self.last_t = None

    def get_clip(self, t1, t2, frame_window_size):
        if self.last_t is not None and t1 < self.last_t:
            raise AssertionError("cannot seek backward")

        vstream = self.vid._container.streams.video[0]
        vs = vstream.start_time * vstream.time_base
        frames = get_frames(
            self.vid._container,
            t1 + vs,
            t2 + vs,
            self.frame_buffer,
            self.frame_buffer_size,
            frame_window_size,
        )
        self.last_t = t1
        return {
            "video": thwc_to_cthw(
                torch.stack(
                    [torch.from_numpy(frame.to_rgb().to_ndarray()) for frame in frames]
                )
            ).to(torch.float32),
            "audio": None,
        }

    @property
    def duration(self) -> float:
        vstream = self.vid._container.streams.video[0]
        return vstream.duration * vstream.time_base


class IndexableVideoDataset(torch.utils.data.Dataset):
    def __init__(
        self,
        config: FeatureExtractConfig,
        videos: List[Video],
        sampler,
        transform,
        frame_window_size,
    ):
        assert (
            config.inference_config.include_audio
            ^ config.inference_config.include_video
        ), """
        cannot include audio and video at the same time
        """
        self.config = config
        self.clips = []
        self.sampler = sampler
        self.transform = transform
        self.frame_window_size = frame_window_size

        if self.config.inference_config.include_video:
            self.encoded_videos = {v.uid: EncodedVideoCached(v.path) for v in videos}
        else:
            assert self.config.inference_config.include_audio
            self.encoded_videos = {
                v.uid: EncodedVideo.from_path(
                    v.path,
                    decode_audio=True,
                    decode_video=False,
                    perform_seek=True,
                    decoder="pyav",
                )
                for v in videos
            }

        for v in videos:
            self.clips.extend(
                list(get_all_clips(v, self.encoded_videos[v.uid].duration, sampler))
            )

    def __len__(self):
        return len(self.clips)

    def __getitem__(self, idx):
        video, clip = self.clips[idx]

        (
            clip_start,
            clip_end,
            clip_index,
            aug_index,
            is_last_clip,
        ) = clip

        encoded_video = self.encoded_videos[video.uid]
        datum = encoded_video.get_clip(clip_start, clip_end, self.frame_window_size)
        v_frames = datum["video"]
        a_frames = datum["audio"]
        sample_dict = {
            "video_name": video.uid,
            "video_index": idx,
            "clip_index": clip_index,
            "aug_index": aug_index,
            "is_stereo": video.is_stereo,
            "clip_start_sec": float(clip_start),
            "clip_end_sec": float(clip_end),
        }
        if v_frames is not None:
            sample_dict["video"] = v_frames
        if a_frames is not None:
            sample_dict["audio"] = a_frames
            sample_dict["audio_sample_rate"] = encoded_video._container.streams.audio[
                0
            ].rate

        sample_dict = self.transform(sample_dict)
        return sample_dict


class CropIfStereo:
    def __init__(self):
        pass

    def __call__(self, x):
        if x["is_stereo"]:
            v = x["video"]
            assert len(v.shape) == 4
            x["video"] = v[:, :, :, 0 : v.shape[-1] // 2]

            # edge case where some videos are incorrectly
            # encoded from source and weren't corrected
            if v.shape[-1] < v.shape[-2]:
                x["video"] = torch.nn.functional.interpolate(
                    x["video"],
                    size=(x["video"].shape[-1], x["video"].shape[-2] // 2),
                    mode="bilinear",
                )
            # for debugging
            # torchvision.utils.save_image(x["video"].permute(1, 0, 2, 3)[0] / 255.0, fp="/tmp/test.jpg")  # noqa
        return x


def get_all_clips(video, video_length, sampler):
    last_clip_time = None
    annotation = {}
    n_clips = 0
    while True:
        clip = sampler(last_clip_time, video_length, annotation)
        last_clip_time = clip.clip_end_sec
        n_clips += 1

        yield (video, clip)

        if clip.is_last_clip:
            break


def labels_fn(
    path: str, start_idx: int, end_idx: int, path_to_video: Dict[str, Any], config
):
    video = path_to_video[path]
    return {
        "video_name": video.uid,
        "is_stereo": video.is_stereo,
        "clip_index": start_idx // config.inference_config.stride,
        "clip_start_sec": start_idx,
        "clip_end_sec": end_idx,
    }


def create_dset(
    videos: List[Video], config: FeatureExtractConfig
) -> IndexableVideoDataset | VideoDataset:
    assert isinstance(videos[0], Video)
    transforms_to_use = [
        CropIfStereo(),
        get_transform(config),
    ]
    if config.io.debug_mode:
        transforms_to_use = [
            CropIfStereo(),
            ApplyTransformToKey(key="video", transform=ShortSideScale(size=256)),
        ]
    transform = Compose(transforms_to_use)

    clip_sampler = UniformClipSampler(
        clip_duration=(
            Fraction(config.inference_config.frame_window, config.inference_config.fps)
            if isinstance(config.inference_config.frame_window, int)
            else config.inference_config.frame_window
        ),
        stride=(
            Fraction(config.inference_config.stride, config.inference_config.fps)
            if isinstance(config.inference_config.stride, int)
            else config.inference_config.stride
        ),
        backpad_last=True,
    )

    return IndexableVideoDataset(
        config,
        videos,
        clip_sampler,
        transform,
        config.inference_config.frame_window,
    )


def create_data_loader(dset, config: FeatureExtractConfig) -> DataLoader:
    if config.inference_config.batch_size == 0:
        raise AssertionError("not supported")

    if config.inference_config.num_workers == -1:  # for debugging
        return dset

    return DataLoader(
        dset,
        batch_size=config.inference_config.batch_size,
        num_workers=config.inference_config.num_workers,
        prefetch_factor=config.inference_config.prefetch_factor,
    )


def create_data_loader_or_dset(
    videos: List[Video], config: FeatureExtractConfig
) -> Any:
    dset = create_dset(videos, config)
    return create_data_loader(dset=dset, config=config)



# Copyright (c) Meta Platforms, Inc. and affiliates. All Rights Reserved.

import functools

import random

import hydra
from ego4d.features.config import FeatureExtractConfig, get_videos, load_model
from ego4d.features.extract_features import extract_features
from ego4d.features.slurm import create_executor


def visualize_extraction(config: FeatureExtractConfig):
    random.seed(1337)

    is_audio_model = config.inference_config.include_audio

    _, videos = get_videos(config)
    if is_audio_model:
        videos = [v for v in videos if v.uid == "046855d7-41d7-4f41-a6a7-fce921ea8133"]
    else:
        videos = [v for v in videos if v.frame_count > 1000 and v.frame_count <= 2000]
    random.shuffle(videos)
    videos = videos[0:5]

    print("Frame count=", videos[0].frame_count)
    print(f"Got {len(videos)} videos")

    config.io.debug_mode = not is_audio_model
    model = load_model(config)

    output_features = extract_features(
        videos=videos,
        config=config,
        model=model,
        log_info=False,
        max_examples=-1,
        silent=True,
        assert_feature_size=not is_audio_model,
    )
    if is_audio_model:
        for video_uid, xs in output_features.result.items():
            print(video_uid)
            for x in xs:
                print(x)
            print()


@hydra.main(config_path="configs", config_name=None)
def schedule_profile_extraction(config: FeatureExtractConfig):
    if config.schedule_config.run_locally:
        visualize_extraction(config)
    else:
        executor = create_executor(config.schedule_config)
        job = executor.submit(functools.partial(visualize_extraction, config=config))
        print(f"{job}")

        # wait for the job
        job.result()


if __name__ == "__main__":
    schedule_profile_extraction()  # pyre-ignore



from types import AuxiliaryVideoComponentDataFile, VideoComponentFile, VideoMetadata

import boto3
import validate
from ffmpeg_utils import VideoInfo

bucket_name = "ego4d-universityaf"
broken_obj = "Participant 4/0249/CMUA_GH010249_04_GOPRO8_16_KNITTING_7_7_2021.MP4"
normal_obj = "Participant 4/0278/CMUA_GH010278_04_GOPRO8_16_KNITTING_7_7_2021.MP4"
uid1 = "123"
uid2 = "456"

metadata = VideoMetadata(
    university_video_id=uid1,
    university_video_folder_path=f"s3://{bucket_name}/Participant 4/0278/",
    number_video_components=2,
    start_date_recorded_utc=None,
    recording_participant_id=0,
    device_id=0,
    video_device_settings={},
    physical_setting_id="0",
    video_scenario_ids=[],
)

component1 = VideoComponentFile(
    university_video_id=uid1,
    video_component_relative_path="CMUA_GH010278_04_GOPRO8_16_KNITTING_7_7_2021.MP4",
    component_index=0,
    is_redacted=False,
    start_date_recorded_utc=None,
    compression_settings={},
    includes_audio=False,
    component_metadata={},
    deidentification_metadata={},
)

component2 = VideoComponentFile(
    university_video_id=uid1,
    video_component_relative_path="2.mp4",
    component_index=2,
    is_redacted=False,
    start_date_recorded_utc=None,
    compression_settings={},
    includes_audio=False,
    component_metadata={},
    deidentification_metadata={},
)

aux1 = AuxiliaryVideoComponentDataFile(
    university_video_id=uid1,
    video_component_relative_path="s3://ego4d-georgiatech/object/key/123/1.mp4",
    component_index=2,
    component_type_id=0,
)

aux2 = AuxiliaryVideoComponentDataFile(
    university_video_id=uid1,
    video_component_relative_path="s3://ego4d-georgiatech/object/key/123/2.mp4",
    component_index=2,
    component_type_id=0,
)

s3 = boto3.client("s3")

aux_dict = {uid1: [aux1, aux2]}


def test_validate_video_components():

    meta_dict = {uid1: metadata}
    component_dict = {uid1: [component1]}

    err = []
    validate._validate_video_components(s3, bucket_name, meta_dict, component_dict, err)
    assert err[0].description == "video_metadata has 1 components when it should have 2"

    component_dict[uid1].append(component2)
    validate._validate_video_components(s3, bucket_name, meta_dict, component_dict, err)
    assert (
        err[1].description
        == "video_metadata has component index 2 when it should have 1"
    )
    assert (
        err[2].description
        == "video s3://ego4d-universityaf/Participant 4/0278/2.mp4 doesn't exist in bucket"
    )

    component_dict = {uid2: [component2, component1]}
    validate._validate_video_components(s3, bucket_name, meta_dict, component_dict, err)
    assert (
        err[3].description
        == "video_component_file points to video_metadata at non-existent university_video_id"
    )


def test_get_videos():
    err = []
    video_info_param = [(uid1, normal_obj), (uid2, broken_obj)]
    video_info_dict = validate._get_videos(s3, bucket_name, video_info_param, err)

    assert (
        err[0].description
        == f"video s3://{bucket_name}/{broken_obj} can't be read by FFMPEG"
    )
    assert len(err) == 1
    assert len(video_info_dict) == 1

    return video_info_dict


def test_validate_mp4(video_info_dict):
    err = []
    validate._validate_mp4(video_info_dict, err)
    err = []
    vinfo = video_info_dict[uid1][0]
    v1 = VideoInfo(
        sample_height=vinfo.sample_height,
        sample_width=vinfo.sample_height + 10,
        vcodec="wrong",
        acodec="wrong",
        sar=vinfo.sar / 2,
        fps=None,
        dar=vinfo.dar / 2,
        mp4_duration=None,
        video_time_base=vinfo.video_time_base / 2,
        vstart=vinfo.vstart,
        astart=vinfo.astart,
        vduration=vinfo.vduration,
        aduration=vinfo.aduration,
    )

    v2 = VideoInfo(
        sample_height=vinfo.sample_height,
        sample_width=vinfo.sample_width,
        vcodec=vinfo.vcodec,
        acodec=vinfo.acodec,
        sar=vinfo.sar,
        fps=vinfo.fps + 1,
        dar=vinfo.dar,
        mp4_duration=0,
        video_time_base=vinfo.video_time_base,
        vstart=vinfo.vstart + 1,
        astart=vinfo.astart,
        vduration=vinfo.vduration,
        aduration=vinfo.aduration,
        rotate=True,
    )

    video_info_dict[uid1][1] = v1
    validate._validate_mp4(video_info_dict, err)

    assert err[0].description == "component 0 has no rotation information"
    assert err[1].description == "component 0 has width < height without rotation"
    assert err[2].description == "component 1 has no rotation information"
    assert err[3].description == "component 1 has null fps value"
    assert err[4].description == "component 1 has no mp4 duration"
    assert err[5].description == "inconsistent video codec"
    assert err[6].description == "components with inconsistent width x height"
    assert err[7].description == "inconsistent video time base"
    assert err[8].description == "inconsistent sar"

    err = []
    video_info_dict[uid1][1] = v2
    validate._validate_mp4(video_info_dict, err)

    assert err[0].description == "component 0 has no rotation information"
    assert err[1].description == "component 0 has width < height without rotation"
    assert err[2].description == "component 1 has mismatching mp4 duration"
    assert err[3].description == "inconsistent video fps"


if __name__ == "__main__":

    test_validate_video_components()
    video_info_dict = test_get_videos()
    video_info_dict[uid1].append(video_info_dict[uid1][0])
    test_validate_mp4(video_info_dict)
